# -*- coding: utf-8 -*-
"""Version related functions and classes are situated here."""

import os
from pathlib import Path
from typing import Any, Dict, List, Optional

import jinja2

from sqlalchemy import Column, ForeignKey, Integer, Table
from sqlalchemy.exc import OperationalError, UnboundExecutionError
from sqlalchemy.orm import Mapped, mapped_column, relationship, synonym, validates

from stalker.db.declarative import Base
from stalker.db.session import DBSession
from stalker.log import get_logger
from stalker.models.entity import Entity
from stalker.models.file import File
from stalker.models.mixins import DAGMixin
from stalker.models.review import Review
from stalker.models.task import Task


logger = get_logger(__name__)


class Version(Entity, DAGMixin):
    """Holds information about the versions created for a class:`.Task`.

    A :class:`.Version` instance holds information about the versions created
    related for a class:`.Task`. This is not directly related to the stored
    files, but instead holds the information about the incremental change
    itself (i.e who has created it, when it is created, the revision and
    version numbers etc.). All the related files are stored in the
    :attr:`.Version.files` attribute.

    .. versionadded: 0.2.13

       After Stalker 0.2.13 the :attr:`.path` become an absolute path which is
       not anymore merged with the project repository in anyway.

    .. warning:

       For projects those are created prior to Stalker version 0.2.13 and that
       has a :class:`.Structure` with :class:`.FilenameTemplate` that doesn't
       include the repository info, it is suggested to update the related
       ``FilenameTemplate`` s to include a the repository info manually.

       Example:
         pre 0.2.13 setup:

         FilenameTemplate with path attribute is set to:

           {{project.code}}/{%- for parent_task in parent_tasks -%}{{parent_task.nice_name}}/{%- endfor -%}

         Update to:

           {{project.repository.path}}/{{project.code}}/{%- for parent_task in parent_tasks -%}{{parent_task.nice_name}}/{%- endfor -%}

         Or, let's have a setup with environment variables:

           $REPO{{project.repository.code}}/{{project.code}}/{%- for parent_task in parent_tasks -%}{{parent_task.nice_name}}/{%- endfor -%}

    .. versionadded: 1.0.0

       Version instances now have an extra numeric counter, preceding the
       :attr:`.version_number` attribute to allow versions to be better
       organized alongside revisions or big changes, without relying on the now
       removed `variant_name` attribute.

    .. versionadded: 1.1.0

       Version class is not deriving from File class anymore. So they are not
       directly related to any file. And the File relation is stored in the new
       :attr:`.Version.files` attribute.

    .. versionadded: 1.1.0

       Added the `files` attribute, which replaces the `outputs` attribute and
       the `inputs` attribute is moved to the :class:`.File` class as the
       `references` attribute, which makes much more sense as individual files
       may reference different `Files` so storing the `references` in `Version`
       doesn't make much sense.

    Args:
        revision_number (Optional[int]): A positive non-zero integer number
            holding the major version counter. This can be set with an
            argument, allowing setting of the revision number as the Version
            instance is created. So, if a :class:`.Version` is created under
            the same :class:`Task` before, the newly created :class:`.Version`
            instances will start from the highest revision number unless it is
            set to another value. Non-sequential revision numbers can be set.
            So, one can start with 1 and then can jump to 3 and 10 from there.
            All the :class:`.Version` instances that have the same
            :attr:`.revision_number` under the same :class:`.Task` will be
            considered in the same version stream and version number attribute
            will be set accordingly. The default is 1.
        files (List[File]): A list of :class:`.File` instances that are created
            for this :class:`.Version` instance. This can be different
            representations (i.e. base, Alembic, USD, ASS, RS etc.) of the same
            data.
        task (Task): A :class:`.Task` instance showing the owner of this
            Version.
        parent (Version): A :class:`.Version` instance which is the parent of
            this Version. It is mainly used to see which Version is derived
            from which in the Version history of a :class:`.Task`.
    """  # noqa: B950

    from stalker import defaults

    __auto_name__ = True
    __tablename__ = "Versions"
    __mapper_args__ = {"polymorphic_identity": "Version"}

    __dag_cascade__ = "save-update, merge"

    version_id: Mapped[int] = mapped_column(
        "id", ForeignKey("Entities.id"), primary_key=True
    )

    __id_column__ = "version_id"

    task_id: Mapped[int] = mapped_column(ForeignKey("Tasks.id"), nullable=False)
    task: Mapped[Task] = relationship(
        primaryjoin="Versions.c.task_id==Tasks.c.id",
        doc="The :class:`.Task` instance that this Version is created for.",
        uselist=False,
        back_populates="versions",
    )

    _revision_number: Mapped[int] = mapped_column(
        "revision_number",
        default=1,
        nullable=False,
    )

    version_number: Mapped[int] = mapped_column(
        default=1,
        nullable=False,
        doc="""The :attr:`.version_number` attribute is read-only.
        Trying to change it will produce an AttributeError.
        """,
    )

    files: Mapped[Optional[List[File]]] = relationship(
        secondary="Version_Files",
        primaryjoin="Versions.c.id==Version_Files.c.version_id",
        secondaryjoin="Version_Files.c.file_id==Files.c.id",
        doc="""The files related to the current version.

        It is a list of :class:`.File` instances.
        """,
    )

    reviews: Mapped[Optional[List[Review]]] = relationship(
        primaryjoin="Reviews.c.version_id==Versions.c.id"
    )

    is_published: Mapped[Optional[bool]] = mapped_column(default=False)

    def __init__(
        self,
        task: Optional[Task] = None,
        files: Optional[List[File]] = None,
        parent: Optional["Version"] = None,
        full_path: Optional[str] = None,
        revision_number: Optional[int] = None,
        **kwargs: Dict[str, Any],
    ) -> None:
        # call supers __init__
        kwargs["full_path"] = full_path
        super(Version, self).__init__(**kwargs)

        DAGMixin.__init__(self, parent=parent)

        self.task = task
        if revision_number is None:
            revision_number = 1
        self.revision_number = revision_number
        self.version_number = None

        if files is None:
            files = []

        self.files = files
        self.is_published = False

    def __repr__(self) -> str:
        """Return the str representation of the Version.

        Returns:
            str: The string representation of this Version instance.
        """
        return (
            "<{project_code}_{nice_name}_v{version_number:03d} ({entity_type})>".format(
                project_code=self.task.project.code,
                nice_name=self.nice_name,
                version_number=self.version_number,
                entity_type=self.entity_type,
            )
        )

    def _validate_revision_number(self, revision_number: int) -> int:
        """Validate the given revision_number value.

        Args:
            revision_number (int): The revision_number value to be validated.

        Raises:
            TypeError: If the given revision_number value is not an integer.
            ValueError: If the given revision_number value is not a positive integer.

        Returns:
            int: The validated revision_number value.
        """
        error_message = (
            f"{self.__class__.__name__}.revision_number should be a "
            f"positive integer, not {revision_number.__class__.__name__}: "
            f"'{revision_number}'"
        )

        if not isinstance(revision_number, int):
            raise TypeError(error_message)

        if revision_number < 1:
            raise ValueError(error_message)

        return revision_number

    def _revision_number_getter(self) -> int:
        """Return the revision_number value.

        Returns:
            int: revision_number attribute value
        """
        return self._revision_number

    def _revision_number_setter(self, revision_number: int):
        """Set the revision attribute value.

        Args:
            revision_number (int): The new revision number value.
        """
        revision_number = self._validate_revision_number(revision_number)

        is_updating_revision_number = False
        if self._revision_number is not None:
            if revision_number != self._revision_number:
                logger.debug(
                    "Updating revision_number from "
                    f"{self._revision_number} -> {revision_number}"
                )
                is_updating_revision_number = True
            else:
                logger.debug(
                    "Revision number is the same... "
                    f"{self._revision_number} == {revision_number}"
                )
        else:
            logger.debug("revision_number is being set for the first time!")

        self._revision_number = revision_number

        if is_updating_revision_number and self.version_number is not None:
            # if we are updating the revision_number value,
            # also update reset the version_number
            logger.debug(
                "Updated revision_number! so, let's update version_number too!"
            )
            logger.debug(f"current version_number is {self.version_number}")
            self.version_number = None

    revision_number: Mapped[int] = synonym(
        "_revision_number",
        descriptor=property(_revision_number_getter, _revision_number_setter),
    )

    @property
    def latest_version(self) -> "Version":
        """Return the Version instance with the highest version number in this series.

        Returns:
            Version: The :class:`.Version` instance with the highest version number in
                this version series.
        """
        latest_version = None
        try:
            with DBSession.no_autoflush:
                latest_version = (
                    Version.query.filter(Version.task == self.task)
                    .filter(Version.revision_number == self.revision_number)
                    .order_by(Version.version_number.desc())
                    .first()
                )
            return latest_version
        except (UnboundExecutionError, OperationalError):
            all_versions = sorted(
                self.task.versions,
                key=lambda x: x.version_number if x.version_number else -1,
            )
            return all_versions[-1] if all_versions else None

    @property
    def max_revision_number(self) -> int:
        """Return the maximum revision number for this Version.

        Returns:
            int: The maximum revision number for this Version.
        """
        with DBSession.no_autoflush:
            result = (
                DBSession.query(Version.revision_number)
                .filter(Version.task_id == self.task_id)
                .order_by(Version.revision_number.desc())
                .first()
            )
        return result[0] if result else 1

    @property
    def max_version_number(self) -> int:
        """Return the maximum version number for this Version.

        Returns:
            int: The maximum version number for this Version.
        """
        latest_version = self.latest_version
        return latest_version.version_number if latest_version else 0

    @validates("version_number")
    def _validate_version_number(self, key: str, version_number: int) -> int:
        """Validate the given version_number value.

        Args:
            key (str): The name of the validated column.
            version_number (int): The version number to be validated.

        Returns:
            int: The validated version number.
        """
        max_version_number = self.max_version_number

        logger.debug(f"max_version_number: {max_version_number}")
        logger.debug(f"given version_number: {version_number}")

        if version_number is not None and version_number > max_version_number:
            return version_number

        if self.latest_version == self:
            if self.version_number is not None:
                version_number = self.version_number
            else:
                version_number = 1
                logger.debug(
                    f"{self.__class__.__name__}.version_number is weirdly 'None', "
                    "no database connection maybe?"
                )
            logger.debug(
                "the version is the latest version in database, the "
                f"number will not be changed from {version_number}"
            )
        else:
            version_number = max_version_number + 1
            logger.debug(
                "given Version.version_number is too low,"
                f"max version_number in the database is {max_version_number}, "
                f"setting the current version_number to {version_number}"
            )

        return version_number

    @validates("task")
    def _validate_task(self, key, task) -> Task:
        """Validate the given task value.

        Args:
            key (str): The name of the validated column.
            task (Task): The task value to be validated.

        Raises:
            TypeError: If the task value is not a :class:`.Task` instance.

        Returns:
            Task: The validated :class:`.Task` instance.
        """
        if task is None:
            raise TypeError("{}.task cannot be None".format(self.__class__.__name__))

        if not isinstance(task, Task):
            raise TypeError(
                "{}.task should be a Task, Asset, Shot, Scene, Sequence or "
                "Variant instance, not {}: '{}'".format(
                    self.__class__.__name__, task.__class__.__name__, task
                )
            )

        return task

    @validates("files")
    def _validate_files(self, key: str, file: File) -> File:
        """Validate the given file value.

        Args:
            key (str): The name of the validated column.
            file (File): The file value to be validated.

        Raises:
            TypeError: If the file is not a :class:`.File` instance.

        Returns:
            File: The validated file value.
        """
        if not isinstance(file, File):
            raise TypeError(
                "{}.files should only contain instances of "
                "stalker.models.file.File, not {}: '{}'".format(
                    self.__class__.__name__, file.__class__.__name__, file
                )
            )

        return file

    def _template_variables(self) -> dict:
        """Return the variables used in rendering the filename template.

        Returns:
            dict: The template variables.
        """
        version_template_vars = {
            "version": self,
            # "extension": self.extension,
        }
        version_template_vars.update(self.task._template_variables())
        return version_template_vars

    def generate_path(self, extension: Optional[str] = None) -> Path:
        """Generate a Path with the template variables from the parent project.

        Args:
            extension (Optional[str]): An optional string containing the
                extension for the resulting Path.

        Raises:
            TypeError: If the extension is not None and not a str.
            RuntimeError: If no Version related FilenameTemplate is found in
                the related `Project.structure`.

        Returns:
            Path: A `pathlib.Path` object.
        """
        if extension is not None and not isinstance(extension, str):
            raise TypeError(
                "extension should be a str, "
                f"not {extension.__class__.__name__}: '{extension}'"
            )
        kwargs = self._template_variables()

        # get a suitable FilenameTemplate
        structure = self.task.project.structure

        vers_template = None
        if structure:
            for template in structure.templates:
                if template.target_entity_type == self.task.entity_type:
                    vers_template = template
                    break

        if not vers_template:
            raise RuntimeError(
                "There are no suitable FilenameTemplate "
                "(target_entity_type == '{entity_type}') defined in the Structure of "
                "the related Project instance, please create a new "
                "stalker.models.template.FilenameTemplate instance with its "
                "'target_entity_type' attribute is set to '{entity_type}' and add "
                "it to the `templates` attribute of the structure of the "
                "project".format(entity_type=self.task.entity_type)
            )

        path = Path(
            jinja2.Template(vers_template.path).render(
                **kwargs, trim_blocks=True, lstrip_blocks=True
            )
        ) / Path(
            jinja2.Template(vers_template.filename).render(
                **kwargs, trim_blocks=True, lstrip_blocks=True
            )
        )
        if extension is not None:
            path = path.with_suffix(extension)

        return path

    @property
    def absolute_full_path(self) -> str:
        """Return the absolute full path of this version.

        This absolute full path includes the repository path of the related project.

        Returns:
            str: The absolute full path of this Version instance.
        """
        return Path(
            os.path.normpath(os.path.expandvars(str(self.generate_path()))).replace(
                "\\", "/"
            )
        )

    @property
    def absolute_path(self) -> str:
        """Return the absolute path.

        Returns:
            str: The absolute path.
        """
        return Path(
            os.path.normpath(
                os.path.expandvars(str(self.generate_path().parent))
            ).replace("\\", "/")
        )

    @property
    def full_path(self) -> Path:
        """Return the full path of this version.

        This full path includes the repository path of the related project as
        it is.

        Returns:
            Path: The full path of this Version instance.
        """
        return self.generate_path()

    @property
    def path(self) -> Path:
        """Return the path.

        Returns:
            Path: The path.
        """
        return self.full_path.parent
    
    @property
    def filename(self) -> str:
        """Return the filename bit of the path.
        
        Returns:
            str: The filename.
        """
        return self.full_path.name

    def is_latest_published_version(self) -> bool:
        """Return True if this is the latest published Version False otherwise.

        Returns:
            bool: True if this is the latest published Version, False otherwise.
        """
        if not self.is_published:
            return False

        return self == self.latest_published_version

    @property
    def latest_published_version(self) -> "Version":
        """Return the last published version.

        Returns:
            Version: The last published Version instance.
        """
        return (
            Version.query.filter_by(task=self.task)
            .filter(Version.revision_number == self.revision_number)
            .filter_by(is_published=True)
            .order_by(Version.version_number.desc())
            .first()
        )

    def __eq__(self, other):
        """Check the equality.

        Args:
            other (object): The other object.

        Returns:
            bool: True if the other object is equal to this one as an Entity, is a
                Version instance, has the same task and same version_number.
        """
        return (
            super(Version, self).__eq__(other)
            and isinstance(other, Version)
            and self.task == other.task
            and self.version_number == other.version_number
        )

    def __hash__(self):
        """Return the hash value of this instance.

        Because the __eq__ is overridden the __hash__ also needs to be overridden.

        Returns:
            int: The hash value.
        """
        return super(Version, self).__hash__()

    @property
    def naming_parents(self) -> List[Task]:
        """Return a list of parents starting from the nearest Asset, Shot or Sequence.

        Returns:
            List[Task]: List of naming parents.
        """
        # find a Asset, Shot or Sequence, and store it as the significant
        # parent, and name the task starting from that entity
        all_parents = self.task.parents
        all_parents.append(self.task)
        naming_parents = []
        if not all_parents:
            return naming_parents

        for parent in reversed(all_parents):
            naming_parents.insert(0, parent)
            if parent.entity_type in ["Asset", "Shot", "Sequence"]:
                break

        return naming_parents

    @property
    def nice_name(self) -> str:
        """Override the nice name method for Version class.

        Returns:
            str: The nice name.
        """
        return self._format_nice_name(
            "_".join(map(lambda x: x.nice_name, self.naming_parents))
        )

    def request_review(self) -> List[Review]:
        """Request a review.

        This is a shortcut to the Task.request_review() method of the related
        task.

        Returns:
            List[Review]: The created Review instances.
        """
        return self.task.request_review(version=self)


# VERSION FILES
Version_Files = Table(
    "Version_Files",
    Base.metadata,
    Column("version_id", Integer, ForeignKey("Versions.id"), primary_key=True),
    Column(
        "file_id",
        Integer,
        ForeignKey("Files.id", onupdate="CASCADE", ondelete="CASCADE"),
        primary_key=True,
    ),
)



# -*- coding: utf-8 -*-

import logging
import pytest

from stalker import log


@pytest.fixture(scope="function")
def setup_logging():
    """Set up stalker log module."""
    log.loggers = []
    yield
    # clean loggers list after every test
    log.loggers = []


def test_register_logger_simple(setup_logging):
    """register logger adds the given logger to the list."""
    logger = logging.getLogger("test_logger")
    assert logger not in log.loggers
    log.register_logger(logger)
    assert logger in log.loggers


def test_register_logger_called_multiple_times(setup_logging):
    """register logger adds the logger only once."""
    logger = logging.getLogger("test_logger")
    assert logger not in log.loggers
    assert 0 == len(log.loggers)
    log.register_logger(logger)
    assert 1 == len(log.loggers)
    log.register_logger(logger)
    assert 1 == len(log.loggers)
    log.register_logger(logger)
    assert 1 == len(log.loggers)
    log.register_logger(logger)
    assert 1 == len(log.loggers)
    assert logger in log.loggers


def test_register_logger_only_accept_loggers(setup_logging):
    """register_logger raise a TypeError if the logger is not a Logger instance."""
    with pytest.raises(TypeError) as cm:
        log.register_logger("not a logger")

    assert str(cm.value) == (
        "logger should be a logging.Logger instance, not str: 'not a logger'"
    )


def test_register_logger_sets_the_level_to_the_default_level(setup_logging):
    """register_logger set the level to the default level."""
    logger = logging.getLogger("logger1")
    logger.setLevel(logging.WARNING)
    assert log.logging_level != logging.WARNING
    log.register_logger(logger)
    assert logger.level == log.logging_level


def test_set_level_sets_all_logger_levels(setup_logging):
    """set_level sets all logger levels all together."""
    logger1 = logging.getLogger("test_logger1")
    logger2 = logging.getLogger("test_logger2")
    logger3 = logging.getLogger("test_logger3")
    logger4 = logging.getLogger("test_logger4")
    logger1.setLevel(logging.DEBUG)
    logger2.setLevel(logging.DEBUG)
    logger3.setLevel(logging.DEBUG)
    logger4.setLevel(logging.DEBUG)
    log.register_logger(logger1)
    log.register_logger(logger2)
    log.register_logger(logger3)
    log.register_logger(logger4)
    assert logger1.level != logging.WARNING
    assert logger2.level != logging.WARNING
    assert logger3.level != logging.WARNING
    assert logger4.level != logging.WARNING
    log.set_level(logging.WARNING)
    assert logger1.level == logging.WARNING
    assert logger2.level == logging.WARNING
    assert logger3.level == logging.WARNING
    assert logger4.level == logging.WARNING


def test_set_level_level_is_not_an_integer(setup_logging):
    """TypeError raised if the logging level is not an integer."""
    with pytest.raises(TypeError) as cm:
        log.set_level("not a logging level")

    assert str(cm.value) == (
        "level should be an integer value one of [0, 10, 20, 30, 40, 50] or "
        "[NOTSET, DEBUG, INFO, WARN, WARNING, ERROR, FATAL, CRITICAL] of the "
        "logging library, not str: 'not a logging level'"
    )


def test_set_level_level_is_not_a_proper_logging_level(setup_logging):
    """ValueError raised if the logging level is not in correct value."""
    with pytest.raises(ValueError) as cm:
        log.set_level(1000)

    assert str(cm.value) == (
        "level should be an integer value one of [0, 10, 20, 30, 40, 50] or "
        "[NOTSET, DEBUG, INFO, WARN, WARNING, ERROR, FATAL, CRITICAL] of the "
        "logging library, not 1000."
    )


def test_get_logger_name_is_not_a_string(setup_logging):
    """stalker.get_logger() raises a TypeError if the name attribute is not
    a str."""
    with pytest.raises(TypeError) as cm:
        log.get_logger(2123)
    assert str(cm.value) == "A logger name must be a string"


def test_get_logger_creates_a_logger(setup_logging):
    """stalker.log.get_logger() returns a Logger instance."""
    logger = log.get_logger("logger")
    assert isinstance(logger, logging.Logger)


def test_get_logger_registers_the_new_logger_already(setup_logging):
    """stalker.log.get_logger() registers the new logger."""
    logger = log.get_logger("logger")
    assert logger in log.loggers


def test_get_logger_sets_the_logging_level_to_the_default_one(setup_logging):
    """stalker.log.get_logger() sets the logging level to the default one."""
    logger = log.get_logger("logger")
    assert logger.level == log.logging_level



# -*- coding: utf-8 -*-
"""Status and StatusList related functions and classes are situated here."""
from typing import Any, Dict, List, Optional, Type, Union

from sqlalchemy import Column, ForeignKey, Integer, Table
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates

from stalker.db.declarative import Base
from stalker.db.session import DBSession
from stalker.log import get_logger
from stalker.models.entity import Entity
from stalker.models.mixins import CodeMixin, TargetEntityTypeMixin

logger = get_logger(__name__)


class Status(Entity, CodeMixin):
    """Defines object statutes.

    No extra parameters, use the *code* attribute to give a short name for the
    status.

    A Status object can be compared with a string value and it will return if
    the lower case name or lower case code of the status matches the lower case
    form of the given string::

    .. code-block:: Python

        >>> from stalker import Status
        >>> a_status = Status(name="On Hold", code="OH")
        >>> a_status == "on hold"
        True
        >>> a_status != "complete"
        True
        >>> a_status == "oh"
        True
        >>> a_status == "another status"
        False

    Args:
        name (str): The name long name of this Status.
        code (str): The code of this Status, its generally the short version of
            the name attribute.
    """

    __auto_name__ = False
    __tablename__ = "Statuses"
    __mapper_args__ = {"polymorphic_identity": "Status"}
    status_id: Mapped[int] = mapped_column(
        "id",
        ForeignKey("Entities.id"),
        primary_key=True,
    )

    def __init__(
        self,
        name: Optional[str] = None,
        code: Optional[str] = None,
        **kwargs: Dict[str, Any],
    ) -> None:
        kwargs["name"] = name
        kwargs["code"] = code

        super(Status, self).__init__(**kwargs)
        self.code = code

    def __eq__(self, other: Any) -> bool:
        """Check the equality.

        Args:
            other (Any): The other object.

        Returns:
            bool: True if the other object is a Status instance and has the same
                attributes.
        """
        if isinstance(other, str):
            return (
                self.name.lower() == other.lower() or self.code.lower() == other.lower()
            )
        else:
            return super(Status, self).__eq__(other) and isinstance(other, Status)

    def __hash__(self) -> int:
        """Return the hash value of this instance.

        Because the __eq__ is overridden the __hash__ also needs to be overridden.

        Returns:
            int: The hash value.
        """
        return super(Status, self).__hash__()


class StatusList(Entity, TargetEntityTypeMixin):
    """Type specific list of :class:`.Status` instances.

    Holds multiple :class:`.Status` instances to be used as a choice list for several
    other classes.

    A StatusList can only be assigned to only one entity type. So a
    :class:`.Project` can only have one suitable StatusList object which is
    designed for :class:`.Project` entities.

    The list of statuses in StatusList can be accessed by using a list like
    indexing and it also supports string indexes only for getting the item,
    you cannot set an item with string indices:

    .. code-block:: Python

        >>> from stalker import Status, StatusList
        >>> status1 = Status(name="Complete", code="CMPLT")
        >>> status2 = Status(name="Work in Progress", code="WIP")
        >>> status3 = Status(name="Pending Review", code="PRev")
        >>> a_status_list = StatusList(name="Asset Status List",
                                    statuses=[status1, status2, status3],
                                    target_entity_type="Asset")
        >>> a_status_list[0]
        <Status (Complete, CMPLT)>
        >>> a_status_list["complete"]
        <Status (Complete, CMPLT)>
        >>> a_status_list["WIP"]
        <Status (Work in Progress, WIP)>

    Args:
        statuses (List[Status]): This is a list of :class:`.Status` instances,
            so you can prepare different StatusLists for different kind of
            entities using the same pool of :class:`.Status` instances.

        target_entity_type (str): use this parameter to specify the target entity
            type that this StatusList is designed for. It accepts classes or names
            of classes.

        For example:

        .. code-block:: Python

            from stalker import Status, StatusList, Project

            status_list = [
                Status(name="Waiting To Start", code="WTS"),
                Status(name="On Hold", code="OH"),
                Status(name="In Progress", code="WIP"),
                Status(name="Waiting Review", code="WREV"),
                Status(name="Approved", code="APP"),
                Status(name="Completed", code="CMPLT"),
            ]

            project_status_list = StatusList(
                name="Project Status List",
                statuses=status_list,
                target_entity_type="Project"
            )

            # or
            project_status_list = StatusList(
                name="Project Status List",
                statuses=status_list,
                target_entity_type=Project
            )

        now with the code above you cannot assign the ``project_status_list``
        object to any other class than a ``Project`` object.

        The StatusList instance can be empty, means it may not have anything in
        its :attr:`.StatusList.statuses`. But it is useless. The validation for
        empty statuses list is left to the SOM user.
    """

    __auto_name__ = True
    __tablename__ = "StatusLists"
    __mapper_args__ = {"polymorphic_identity": "StatusList"}

    __unique_target__ = True

    status_list_id: Mapped[int] = mapped_column(
        "id",
        ForeignKey("Entities.id"),
        primary_key=True,
    )

    statuses: Mapped[Optional[List[Status]]] = relationship(
        secondary="StatusList_Statuses",
        doc="List of :class:`.Status` objects, showing the possible statuses",
    )

    def __init__(
        self,
        statuses: Optional[List[Status]] = None,
        target_entity_type: Optional[Union[Type, str]] = None,
        **kwargs: Dict[str, Any],
    ) -> None:
        super(StatusList, self).__init__(**kwargs)
        TargetEntityTypeMixin.__init__(self, target_entity_type, **kwargs)

        if statuses is None:
            statuses = []
        self.statuses = statuses

    @validates("statuses")
    def _validate_statuses(self, key: str, status: Status) -> Status:
        """Validate the given status value.

        Args:
            key (str): The name of the validated column.
            status (Status): The status value to be validated.

        Raises:
            TypeError: If the status value is not a Status instance.

        Returns:
            Status: The validated status value.
        """
        if not isinstance(status, Status):
            raise TypeError(
                f"All of the elements in {self.__class__.__name__}.statuses must be an "
                "instance of stalker.models.status.Status, "
                f"not {status.__class__.__name__}: '{status}'"
            )
        return status

    def __eq__(self, other: Any) -> bool:
        """Check the equality.

        Args:
            other (Any): The other object.

        Returns:
            bool: True if the other object is a StatusList instance and has the same
                statuses, target_entity_type.
        """
        return (
            super(StatusList, self).__eq__(other)
            and isinstance(other, StatusList)
            and self.statuses == other.statuses
            and self.target_entity_type == other.target_entity_type
        )

    def __hash__(self) -> int:
        """Return the hash value of this instance.

        Because the __eq__ is overridden the __hash__ also needs to be overridden.

        Returns:
            int: The hash value.
        """
        return super(StatusList, self).__hash__()

    def __getitem__(self, key: int) -> Status:
        """Return the Status at the given key.

        Args:
            key (int): The index to return the value of.

        Returns:
            Status: The Status instance at the given index.
        """
        return_item = None
        with DBSession.no_autoflush:
            if isinstance(key, str):
                for item in self.statuses:
                    if item == key:
                        return_item = item
                        break
            else:
                return_item = self.statuses[key]

        return return_item

    def __setitem__(self, key: int, value: Status) -> None:
        """Set the value at the given index.

        Args:
            key (int): The index to set the item value to.
            value (Status): The Status instance to set at the given index.
        """
        self.statuses[key] = value

    def __delitem__(self, key: int) -> None:
        """Delete the item with the given key.

        Args:
            key (int): Remove the Status at the given index.
        """
        del self.statuses[key]

    def __len__(self) -> int:
        """Return the  number of Statuses in this StatusList.

        Returns:
            int: The number of Statuses in this StatusList.
        """
        return len(self.statuses)


# StatusList_Statuses Table
StatusList_Statuses = Table(
    "StatusList_Statuses",
    Base.metadata,
    Column("status_list_id", Integer, ForeignKey("StatusLists.id"), primary_key=True),
    Column("status_id", Integer, ForeignKey("Statuses.id"), primary_key=True),
)



# -*- coding: utf-8 -*-
"""Project related classes and functions are situated here."""

from typing import Any, List, Optional, TYPE_CHECKING, Union

from sqlalchemy import Float, ForeignKey
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.orderinglist import ordering_list
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates

from stalker.db.declarative import Base
from stalker.db.session import DBSession
from stalker.log import get_logger
from stalker.models.entity import Entity
from stalker.models.mixins import CodeMixin, DateRangeMixin, ReferenceMixin, StatusMixin
from stalker.models.status import Status

if TYPE_CHECKING:  # pragma: no cover
    from stalker.models.asset import Asset
    from stalker.models.auth import Role, User
    from stalker.models.client import Client
    from stalker.models.format import ImageFormat
    from stalker.models.repository import Repository
    from stalker.models.sequence import Sequence
    from stalker.models.shot import Shot
    from stalker.models.structure import Structure
    from stalker.models.task import Task
    from stalker.models.ticket import Ticket

logger = get_logger(__name__)


class ProjectRepository(Base):
    """The association object for Project to Repository instances."""

    __tablename__ = "Project_Repositories"

    project_id: Mapped[int] = mapped_column(
        ForeignKey("Projects.id"),
        primary_key=True,
    )

    project: Mapped["Project"] = relationship(
        back_populates="repositories_proxy",
        primaryjoin="Project.project_id==ProjectRepository.project_id",
    )

    repository_id: Mapped[int] = mapped_column(
        ForeignKey("Repositories.id"),
        primary_key=True,
    )

    repository: Mapped["Repository"] = relationship(
        primaryjoin="ProjectRepository.repository_id==Repository.repository_id",
    )

    position: Mapped[Optional[int]] = mapped_column()

    def __init__(
        self,
        project: Optional["Project"] = None,
        repository: Optional["Repository"] = None,
        position: Optional[int] = None,
    ) -> None:
        self.project = project
        self.repository = repository
        self.position = position

    @validates("project")
    def _validate_project(self, key: str, project: "Project") -> "Project":
        """Validate the given project value.

        Args:
            key (str): The name of the validated column.
            project (Project): The project value to be validated.

        Returns:
            project: The validated project value.
        """
        # TODO: Why we are not validating the Project here?
        #       Is it already validated somewhere else?
        return project

    @validates("repository")
    def _validate_repository(
        self,
        key: str,
        repository: Union[None, "Repository"],
    ) -> Union[None, "Repository"]:
        """Validate the given repository value.

        Args:
            key (str): The name of the validated column.
            repository (Repository): The repository to be validated.

        Raises:
            TypeError: If the repository is not a Repository instance.

        Returns:
            Repository: The repository value.
        """
        if repository is not None:
            from stalker.models.repository import Repository

            if not isinstance(repository, Repository):
                raise TypeError(
                    f"{self.__class__.__name__}.repositories should be a list of "
                    "stalker.models.repository.Repository instances or "
                    f"derivatives, not {repository.__class__.__name__}: '{repository}'"
                )

        return repository


class Project(Entity, ReferenceMixin, StatusMixin, DateRangeMixin, CodeMixin):
    """All the information about a Project in Stalker is hold in this class.

    Project is one of the main classes that will direct the others. A project
    in Stalker is a gathering point.

    It is mixed with :class:`.ReferenceMixin`, :class:`.StatusMixin`,
    :class:`.DateRangeMixin` and :class:`.CodeMixin` to give reference, status,
    schedule and code attribute. Please read the individual documentation of
    each of the mixins.

    **Project Users**

    The :attr:`.Project.users` attribute lists the users in this project. UIs
    like task creation for example will only list these users as available
    resources for this project.

    **TaskJuggler Integration**

    Stalker uses TaskJuggler for scheduling the project tasks. The
    :attr:`.Project.to_tjp` attribute generates a tjp compliant string which
    includes the project definition, the tasks of the project, the resources in
    the project including the vacation definitions and all the time logs
    recorded for the project.

    For custom attributes or directives that needs to be passed to TaskJuggler
    you can use the :attr:`.Project.custom_tjp` attribute which will be
    attached to the generated tjp file (inside the "project" directive).

    To manage all the studio projects at once (schedule them at once please use
    :class:`.Studio`).

    **Repositories**

    .. versionadded:: 0.2.13
       Multiple Repositories per Project

       Starting with v0.2.13 Project instances can have multiple Repositories,
       which allows the project files to be placed in more than one repository
       according to the need of the studio pipeline. One great advantage of
       having multiple repositories is to be able to place Published versions
       in to another repository which is placed on to a faster server.

       Also, the :attr:`.repositories` attribute is not a read-only attribute
       anymore.

    **Clients**

    .. versionadded:: 0.2.15
       Multiple Clients per Project

       It is now possible to attach multiple :class:`.Client` instances to one
       :class:`.Project` allowing to hold complex Projects to Client relations
       by using the :attr:`.ProjectClient.role` attribute of the
       :class:`.ProjectClient` class.

    **Deleting a Project**

    Deleting a :class:`.Project` instance will cascade the delete operation to
    all the :class:`.Task` s related to that particular Project and it will
    cascade the delete operation to :class:`.TimeLog` s, :class:`.Version` s,
    :class:`.File` s and :class:`.Review` s etc.. So one can delete a
    :class:`.Project` instance without worrying about the non-project related
    data like :class:`.User` s or :class:`.Department` s to be deleted.

    Args:
        clients (List[Client]): The clients which the project is affiliated with.
            Default value is an empty list.

        image_format (ImageFormat): The output image format of the project. Default
            value is None.

        fps (float): The FPS of the project, it should be a integer or float number, or
            a string literal which can be correctly converted to a float. Default value
            is 25.0.

        type (Type): The type of the project. Default value is None.

        structure (Structure): The structure of the project. Default value is None.

        repositories (List[Repository]): A list of :class:`.Repository` instances that
            the project files are going to be stored in. You cannot create a project
            without specifying the repositories argument and passing a
            :class:`.Repository` to it. Default value is None which raises a TypeError.

        is_stereoscopic (bool): a bool value, showing if the project is going to be a
            stereo 3D project, anything given as the argument will be converted to True
            or False. Default value is False.

        users (List[User]): A list of :class:`.User` s holding the users in this
            project. This will create a reduced or grouped list of studio workers and
            will make it easier to define the resources for a Task related to this
            project. The default value is an empty list.
    """

    __auto_name__ = False
    __tablename__ = "Projects"
    project_id: Mapped[int] = mapped_column(
        "id",
        ForeignKey("Entities.id"),
        primary_key=True,
    )

    __mapper_args__ = {
        "polymorphic_identity": "Project",
        "inherit_condition": project_id == Entity.entity_id,
    }

    clients = association_proxy(
        "client_role", "client", creator=lambda n: ProjectClient(client=n)
    )

    client_role: Mapped[Optional[List["ProjectClient"]]] = relationship(
        back_populates="project",
        cascade="all, delete-orphan",
        cascade_backrefs=False,
        primaryjoin="Projects.c.id==Project_Clients.c.project_id",
    )

    tasks: Mapped[Optional[List["Task"]]] = relationship(
        primaryjoin="Tasks.c.project_id==Projects.c.id",
        cascade="all, delete-orphan",
    )

    users = association_proxy(
        "user_role", "user", creator=lambda n: ProjectUser(user=n)
    )

    user_role: Mapped[Optional[List["ProjectUser"]]] = relationship(
        back_populates="project",
        cascade="all, delete-orphan",
        cascade_backrefs=False,
        primaryjoin="Projects.c.id==Project_Users.c.project_id",
    )

    repositories_proxy: Mapped[Optional[List["ProjectRepository"]]] = relationship(
        back_populates="project",
        cascade="all, delete-orphan",
        cascade_backrefs=False,
        order_by="ProjectRepository.position",
        primaryjoin="Projects.c.id==Project_Repositories.c.project_id",
        collection_class=ordering_list("position"),
        doc="""The :class:`.Repository` that this project files should reside.

        Should be a list of :class:`.Repository` instances.
        """,
    )

    repositories = association_proxy(
        "repositories_proxy",
        "repository",
        creator=lambda n: ProjectRepository(repository=n),
    )

    structure_id: Mapped[Optional[int]] = mapped_column(ForeignKey("Structures.id"))
    structure: Mapped[Optional["Structure"]] = relationship(
        primaryjoin="Project.structure_id==Structure.structure_id",
        doc="""The structure of the project. Should be an instance of
        :class:`.Structure` class""",
    )

    image_format_id: Mapped[Optional[int]] = mapped_column(
        ForeignKey("ImageFormats.id")
    )
    image_format: Mapped[Optional["ImageFormat"]] = relationship(
        primaryjoin="Projects.c.image_format_id==ImageFormats.c.id",
        doc="""The :class:`.ImageFormat` of this project.

        This value defines the output image format of the project, should be an
        instance of :class:`.ImageFormat`.
        """,
    )

    fps: Mapped[Optional[float]] = mapped_column(
        Float(precision=3),
        doc="""The fps of the project.

        It is a float value, any other types will be converted to float. The
        default value is 25.0.
        """,
    )

    is_stereoscopic: Mapped[Optional[bool]] = mapped_column(
        doc="""True if the project is a stereoscopic project"""
    )

    tickets: Mapped[Optional[List["Ticket"]]] = relationship(
        primaryjoin="Tickets.c.project_id==Projects.c.id",
        cascade="all, delete-orphan",
    )

    def __init__(
        self,
        name: Optional[str] = None,
        code: Optional[str] = None,
        clients: Optional[List["Client"]] = None,
        repositories: Optional[List["Repository"]] = None,
        structure: Optional["Structure"] = None,
        image_format: Optional["ImageFormat"] = None,
        fps: float = 25.0,
        is_stereoscopic: bool = False,
        users: Optional[List["User"]] = None,
        **kwargs,
    ) -> None:
        # a projects project should be self
        # initialize the project argument to self
        kwargs["project"] = self
        kwargs["name"] = name

        super(Project, self).__init__(**kwargs)
        # call the mixin __init__ methods
        ReferenceMixin.__init__(self, **kwargs)
        StatusMixin.__init__(self, **kwargs)
        DateRangeMixin.__init__(self, **kwargs)

        self.code = code

        if users is None:
            users = []
        self.users = users

        if repositories is None:
            repositories = []
        self.repositories = repositories

        self.structure = structure

        if clients is None:
            clients = []
        self.clients = clients

        self._sequences = []
        self._assets = []

        self.image_format = image_format
        self.fps = fps
        self.is_stereoscopic = bool(is_stereoscopic)

    def __eq__(self, other: Any) -> bool:
        """Check the equality.

        Args:
            other (Any): The other object.

        Returns:
            bool: True if the other object is a Project and equal as an Entity.
        """
        return super(Project, self).__eq__(other) and isinstance(other, Project)

    def __hash__(self) -> int:
        """Return the hash value of this instance.

        Because the __eq__ is overridden the __hash__ also needs to be overridden.

        Returns:
            int: The hash value.
        """
        return super(Project, self).__hash__()

    @validates("fps")
    def _validate_fps(self, key: str, fps: Union[int, float]) -> float:
        """Validate the given fps value.

        Args:
            key (str): The name of the validated column.
            fps (Union[int, float]): The fps value to be validated.

        Raises:
            TypeError: If the fps value is not an int or float.
            ValueError: If the fps is 0 or a negative number.

        Returns:
            float: The validated fps value.
        """
        if not isinstance(fps, (int, float)):
            raise TypeError(
                f"{self.__class__.__name__}.fps should be a positive float or int, "
                f"not {fps.__class__.__name__}: '{fps}'"
            )

        fps = float(fps)
        if fps <= 0:
            raise ValueError(
                f"{self.__class__.__name__}.fps should be a positive float or int, "
                f"not {fps}"
            )
        return float(fps)

    @validates("image_format")
    def _validate_image_format(
        self, key: str, image_format: Union[None, "ImageFormat"]
    ) -> Union[None, "ImageFormat"]:
        """Validate the given image format.

        Args:
            key (str): The name of the validated column.
            image_format (Union[None, ImageFormat]): The image_format value to
                be validated.

        Raises:
            TypeError: If the given image format is not a ImageFormat instance.

        Returns:
            Union[None, ImageFormat]: The validated image_format value.
        """
        from stalker.models.format import ImageFormat

        if image_format is not None and not isinstance(image_format, ImageFormat):
            raise TypeError(
                f"{self.__class__.__name__}.image_format should be an instance of "
                "stalker.models.format.ImageFormat, "
                f"not {image_format.__class__.__name__}: '{image_format}'"
            )
        return image_format

    @validates("structure")
    def _validate_structure(
        self,
        key: str,
        structure: Union[None, "Structure"],
    ) -> Union[None, "Structure"]:
        """Validate the given structure value.

        Args:
            key (str): The name of the validated column.
            structure (Structure): The structure to be validated.

        Raises:
            TypeError: If the given structure is not a Structure instance.

        Returns:
            Structure: The validated Structure value.
        """
        from stalker.models.structure import Structure

        if structure is not None and not isinstance(structure, Structure):
            raise TypeError(
                "{}.structure should be an instance of "
                "stalker.models.structure.Structure, not {}: '{}'".format(
                    self.__class__.__name__, structure.__class__.__name__, structure
                )
            )
        return structure

    @validates("is_stereoscopic")
    def _validate_is_stereoscopic(
        self,
        key: str,
        is_stereoscopic: bool,
    ) -> bool:
        """Validate the is_stereoscopic value.

        Args:
            key (str): The name of the validated column.
            is_stereoscopic (bool): The is_stereoscopic value to be validated.

        Returns:
            bool: The bool representation of the is_stereoscopic value.
        """
        return bool(is_stereoscopic)

    @property
    def root_tasks(self) -> List["Task"]:
        """Return a list of Tasks which have no parents.

        Returns:
            List[Task]: The list of root :class:`Task`s in this project.
        """
        from stalker.models.task import Task
        from stalker.db.session import DBSession

        # TODO: add a fallback method
        with DBSession.no_autoflush:
            return (
                Task.query.filter(Task.project == self)
                .filter(Task.parent == None)  # noqa: E711
                .all()
            )

    @property
    def assets(self) -> List["Asset"]:
        """Return the assets in this project.

        Returns:
            List[Asset]: The list of :class:`Asset`s in this project.
        """
        from stalker.models.asset import Asset
        from stalker.db.session import DBSession

        # TODO: add a fallback method
        with DBSession.no_autoflush:
            return Asset.query.filter(Asset.project == self).all()

    @property
    def sequences(self) -> List["Sequence"]:
        """Return the sequences in this project.

        Returns:
            List[Sequence]: List of :class:`Sequence`s in this project.
        """
        # sequences are tasks, use self.tasks
        from stalker.models.sequence import Sequence

        return Sequence.query.filter(Sequence.project == self).all()

    @property
    def shots(self) -> List["Shot"]:
        """Return the shots in this project.

        Returns:
            List[Shot]: List of :class:`Shot`s in this project.
        """
        # shots are tasks, use self.tasks
        from stalker.models.shot import Shot

        return Shot.query.filter(Shot.project == self).all()

    @property
    def to_tjp(self) -> str:
        """Return the TaskJuggler compatible representation of this project.

        Returns:
            str: The TaskJuggler compatible representation of this project.
        """
        tab = "    "
        indent = tab
        tjp = f'task {self.tjp_id} "{self.tjp_id}" {{'
        for task in self.root_tasks:
            tjp += "\n"
            tjp += "\n".join(f"{indent}{line}" for line in task.to_tjp.split("\n"))

        tjp += "\n}"
        return tjp

    @property
    def is_active(self) -> bool:
        """Return True if this project is active, False otherwise.

        This is a predicate for `Project.active` attribute.

        Returns:
            bool: True if the project is active, False otherwise.
        """
        with DBSession.no_autoflush:
            wip = Status.query.filter_by(code="WIP").first()
        return self.status == wip

    @property
    def total_logged_seconds(self) -> int:
        """Return the total TimeLog seconds recorded in child tasks.

        Returns:
            int: The total amount of logged seconds in the child tasks.
        """
        total_logged_seconds = 0
        for task in self.root_tasks:
            total_logged_seconds += task.total_logged_seconds
        logger.debug(f"project.total_logged_seconds: {total_logged_seconds}")
        return total_logged_seconds

    @property
    def schedule_seconds(self) -> int:
        """Return the total amount of schedule timing of the child tasks in seconds.

        Returns:
            int: The total amount of schedule timing of the child tasks in seconds.
        """
        schedule_seconds = 0
        for task in self.root_tasks:
            schedule_seconds += task.schedule_seconds
        logger.debug(f"project.schedule_seconds: {schedule_seconds}")
        return schedule_seconds

    @property
    def percent_complete(self) -> float:
        """Return the percent_complete value.

        The percent_complete value is based on the total_logged_seconds and
        schedule_seconds of the root tasks.

        Returns:
            float: The percent_complete value.
        """
        total_logged_seconds = self.total_logged_seconds
        schedule_seconds = self.schedule_seconds
        if schedule_seconds > 0:
            return total_logged_seconds / schedule_seconds * 100
        else:
            return 0

    @property
    def open_tickets(self) -> List["Ticket"]:
        """Return the list of open :class:`.Ticket` s in this project.

        Returns:
             List[Ticket]: A list of :class:`.Ticket` instances which has a status of
                `Open` and created in this project.
        """
        from stalker import Ticket, Status

        return (
            Ticket.query.join(Status, Ticket.status)
            .filter(Ticket.project == self)
            .filter(Status.code != "CLS")
            .all()
        )

    @property
    def repository(self) -> "Repository":
        """Return the first repository in the `project.repositories` or None.

        Compatibility attribute for pre v0.2.13 systems.

        Returns:
            Union[None, Repository]: The Repository instance if there are any or None.
        """
        if self.repositories:
            return self.repositories[0]
        else:
            return None


class ProjectUser(Base):
    """The association object used in User-to-Project relation."""

    __tablename__ = "Project_Users"

    user_id: Mapped[int] = mapped_column(
        "user_id", ForeignKey("Users.id"), primary_key=True
    )

    user: Mapped["User"] = relationship(
        back_populates="project_role",
        cascade_backrefs=False,
        primaryjoin="ProjectUser.user_id==User.user_id",
    )

    project_id: Mapped[int] = mapped_column(ForeignKey("Projects.id"), primary_key=True)

    project: Mapped[Project] = relationship(
        back_populates="user_role",
        cascade_backrefs=False,
        primaryjoin="ProjectUser.project_id==Project.project_id",
    )

    role_id: Mapped[Optional[int]] = mapped_column("rid", ForeignKey("Roles.id"))

    role: Mapped[Optional["Role"]] = relationship(
        "Role", cascade_backrefs=False, primaryjoin="ProjectUser.role_id==Role.role_id"
    )

    rate: Mapped[Optional[float]] = mapped_column(default=0.0)

    def __init__(
        self,
        project: Optional[Project] = None,
        user: Optional["User"] = None,
        role: Optional["Role"] = None,
    ) -> None:
        self.user = user
        self.project = project
        self.role = role
        if self.user:
            # don't need to validate rate
            # as it is already validated on the User side
            self.rate = user.rate

    @validates("user")
    def _validate_user(
        self,
        key: str,
        user: Union[None, "User"],
    ) -> Union[None, "User"]:
        """Validate the given user value.

        Args:
            key (str): The name of the validated column.
            user (User): The user value to be validated.

        Raises:
            TypeError: If the given user is not a User instance.

        Returns:
            User: The validated user value.
        """
        if user is not None:
            from stalker.models.auth import User

            if not isinstance(user, User):
                raise TypeError(
                    f"{self.__class__.__name__}.user should be a "
                    "stalker.models.auth.User instance, "
                    f"not {user.__class__.__name__}: '{user}'"
                )

            # also update rate attribute
            from stalker.db.session import DBSession

            with DBSession.no_autoflush:
                self.rate = user.rate

        return user

    @validates("project")
    def _validate_project(
        self, key: str, project: Union[None, Project]
    ) -> Union[None, Project]:
        """Validate the given project value.

        Args:
            key (str): The name of the validated column.
            project (Union[None, Project]): The project value to be validated.

        Raises:
            TypeError: If the project is not a Project instance.

        Returns:
            Union[None, Project]: The validated project value.
        """
        if project is not None:
            # check if it is instance of Project object
            if not isinstance(project, Project):
                raise TypeError(
                    f"{self.__class__.__name__}.project should be a "
                    "stalker.models.project.Project instance, "
                    f"not {project.__class__.__name__}: '{project}'"
                )
        return project

    @validates("role")
    def _validate_role(self, key: str, role: Union[None, "Role"]):
        """Validate the given role instance.

        Args:
            key (str): The name of the validated column.
            role (Union[None, "Role"]): The role value to be validated.

        Raises:
            TypeError: If the given role is not a Role instance.

        Returns:
            Union[None, "Role"]: The validated role value.
        """
        if role is None:
            return role

        from stalker import Role

        if not isinstance(role, Role):
            raise TypeError(
                f"{self.__class__.__name__}.role should be a "
                "stalker.models.auth.Role instance, "
                f"not {role.__class__.__name__}: '{role}'"
            )

        return role


class ProjectClient(Base):
    """The association object used in Client-to-Project relation.

    Args:
        project (Project): The project.
        client (Client): The client.
        role (Role): The client role in this project.
    """

    __tablename__ = "Project_Clients"

    client_id: Mapped[int] = mapped_column(ForeignKey("Clients.id"), primary_key=True)

    client: Mapped["Client"] = relationship(
        back_populates="project_role",
        cascade_backrefs=False,
        primaryjoin="Project_Clients.c.client_id==Clients.c.id",
    )

    project_id: Mapped[int] = mapped_column(ForeignKey("Projects.id"), primary_key=True)

    project: Mapped[Project] = relationship(
        back_populates="client_role",
        cascade_backrefs=False,
        primaryjoin="ProjectClient.project_id==Project.project_id",
    )

    role_id: Mapped[Optional[int]] = mapped_column(
        "rid", ForeignKey("Roles.id"), nullable=True
    )

    role: Mapped[Optional["Role"]] = relationship(
        cascade_backrefs=False,
        primaryjoin="ProjectClient.role_id==Role.role_id",
    )

    def __init__(
        self,
        project: Optional[Project] = None,
        client: Optional["Client"] = None,
        role: Optional["Role"] = None,
    ) -> None:
        self.client = client
        self.project = project
        self.role = role

    @validates("client")
    def _validate_client(
        self,
        key: str,
        client: Union[None, "Client"],
    ) -> Union[None, "Client"]:
        """Validate the given client value.

        Args:
            key (str): The name of the validated column.
            client (Union[None, Client]): The client value to be validated.

        Raises:
            TypeError: If the given client arg value is not a Client instance.

        Returns:
            Client: The validated client value.
        """
        if client is None:
            return client

        from stalker.models.client import Client

        if not isinstance(client, Client):
            raise TypeError(
                f"{self.__class__.__name__}.client should be an instance of "
                "stalker.models.auth.Client, "
                f"not {client.__class__.__name__}: '{client}'"
            )

        return client

    @validates("project")
    def _validate_project(
        self, key: str, project: Union[None, Project]
    ) -> Union[None, Project]:
        """Validate the given project value.

        Args:
            key (str): The name of the validated column.
            project (Project): The project value to be validated.

        Raises:
            TypeError: If the given project value is not a Project instance.

        Returns:
            Project: The validated project value.
        """
        if project is None:
            return project

        # check if it is instance of Project object
        if not isinstance(project, Project):
            raise TypeError(
                f"{self.__class__.__name__}.project should be a "
                "stalker.models.project.Project instance, "
                f"not {project.__class__.__name__}: '{project}'"
            )

        return project

    @validates("role")
    def _validate_role(
        self,
        key: str,
        role: Union[None, "Role"],
    ) -> Union[None, "Role"]:
        """Validate the given role instance.

        Args:
            key (str): The name of the validated column.
            role (Union[None, Role]): The role value to be validated.

        Raises:
            TypeError: If the given role value is not a Role instance.

        Returns:
            Union[None, Role]: The validated role value.
        """
        if role is None:
            return role

        from stalker import Role

        if not isinstance(role, Role):
            raise TypeError(
                f"{self.__class__.__name__}.role should be a "
                "stalker.models.auth.Role instance, "
                f"not {role.__class__.__name__}: '{role}'"
            )

        return role


def create_project_client(project: Project) -> ProjectClient:
    """Create ProjectClient instance on association proxy.

    Args:
        project (Project): The :class:`.Project` instance to be used to create the
            :class:`.ProjectClient` instance.

    Returns:
        ProjectClient: The :class:`.ProjectClient` instance.
    """
    return ProjectClient(project=project)



# -*- coding: utf-8 -*-
"""Budget related classes and functions are situated here."""
from typing import Any, Dict, List, Optional, TYPE_CHECKING, Union

from sqlalchemy import Column, Float, ForeignKey, Integer, String, Table
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates

from stalker.db.declarative import Base
from stalker.models.entity import Entity
from stalker.models.mixins import (
    AmountMixin,
    DAGMixin,
    ProjectMixin,
    StatusMixin,
    UnitMixin,
)

if TYPE_CHECKING:  # pragma: no cover
    from stalker.models.client import Client


class Good(Entity, UnitMixin):
    """Manages commercial items that is served by the Studio.

    A Studio can define service prices or items that's been sold by the Studio
    by using a list of commercial items.

    .. note::
       .. versionadded 0.2.20: Client Specific Goods

       Clients now can own a list of :class:`.Good` s attached to them.
       So one can define a list of :class:`.Good` s with special prices
       adjusted for a particular ``Client``, then get them back from the db by
       querying the :class:`.Good` s those have their ``client`` attribute set
       to that particular ``Client`` instance. Removing a ``Good`` from a
       :class:`.Client` will not delete it from the database, but deleting a
       :class:`.Client` will also delete the ``Good`` s attached to that
       particular :class:`.Client`.

    .. ::
       don't forget to update the Client documentation, which also has the same
       text.

    A Good has the following attributes

    Args:
        cost (Union[int, float]): The cost of this item to the Studio, so
            generally it is better to keep price of the related BudgetEntry
            bigger than this value to get profit by selling this item.
        msrp (Union[int, float]): The suggested retail price for this item.
        unit (str): The unit of this item.
    """

    __auto_name__ = False
    __tablename__ = "Goods"
    __mapper_args__ = {"polymorphic_identity": "Good"}

    good_id: Mapped[int] = mapped_column(
        "id", ForeignKey("Entities.id"), primary_key=True
    )

    price_lists: Mapped["PriceList"] = relationship(
        secondary="PriceList_Goods",
        primaryjoin="Goods.c.id==PriceList_Goods.c.good_id",
        secondaryjoin="PriceList_Goods.c.price_list_id==PriceLists.c.id",
        back_populates="goods",
        doc="PriceLists that this good is related to.",
    )

    cost: Mapped[Optional[float]] = mapped_column(default=0.0)
    msrp: Mapped[Optional[float]] = mapped_column(Float, default=0.0)
    unit: Mapped[Optional[str]] = mapped_column(String(64))

    client_id: Mapped[Optional[int]] = mapped_column(
        "client_id", ForeignKey("Clients.id")
    )
    client: Mapped["Client"] = relationship(
        primaryjoin="Goods.c.client_id==Clients.c.id",
        back_populates="goods",
        uselist=False,
    )

    def __init__(
        self,
        cost: Union[int, float] = 0.0,
        msrp: Union[int, float] = 0.0,
        unit: str = "",
        client: Optional["Client"] = None,
        **kwargs,
    ) -> None:
        super(Good, self).__init__(**kwargs)
        UnitMixin.__init__(self, unit=unit)
        self.cost = cost
        self.msrp = msrp
        self.client = client

    @validates("cost")
    def _validate_cost(self, key: str, cost: Union[int, float]) -> Union[int, float]:
        """Validate the given cost value.

        Args:
            key (str): The name of the validated column.
            cost (Union[int, float]): The cost value to be validated.

        Raises:
            TypeError: If the given cost value is not an int or float.
            ValueError: If the given cost is a negative value.

        Returns:
            float: The validated cost value.
        """
        if cost is None:
            cost = 0.0

        if not isinstance(cost, (float, int)):
            raise TypeError(
                f"{self.__class__.__name__}.cost should be a non-negative number, "
                f"not {cost.__class__.__name__}: '{cost}'"
            )

        if cost < 0.0:
            raise ValueError(
                f"{self.__class__.__name__}.cost should be a non-negative number"
            )

        return cost

    @validates("msrp")
    def _validate_msrp(self, key: str, msrp: Union[int, float]) -> Union[int, float]:
        """Validate the given msrp value.

        Args:
            key (str): The name of the validated column.
            msrp (Union[int, float]): The msrp value to be validated.

        Raises:
            TypeError: If the given msrp value is not an int or float.
            ValueError: If the msrp is a negative value.

        Returns:
            float: The validated msrp value.
        """
        if msrp is None:
            msrp = 0.0

        if not isinstance(msrp, (float, int)):
            raise TypeError(
                f"{self.__class__.__name__}.msrp should be a non-negative number, "
                f"not {msrp.__class__.__name__}: '{msrp}'"
            )

        if msrp < 0.0:
            raise ValueError(
                f"{self.__class__.__name__}.msrp should be a non-negative number"
            )

        return msrp

    @validates("client")
    def _validate_client(self, key: str, client: "Client") -> "Client":
        """Validate the given client value.

        Args:
            key (str): The name of the validated column.
            client (Client): The client value to be validated.

        Raises:
            TypeError: If the given client arg value is not a
                :class:`stalker.models.client.Client` instance.

        Returns:
            Client: The validated :class:`stalker.models.client.Client` instance.
        """
        if client is not None:
            from stalker.models.client import Client

            if not isinstance(client, Client):
                raise TypeError(
                    f"{self.__class__.__name__}.client attribute should be a "
                    "stalker.models.client.Client instance, "
                    f"not {client.__class__.__name__}: '{client}'"
                )
        return client


class PriceList(Entity):
    """Contains CommercialItems to create a list of items that is sold by the Studio.

    You can create different lists for items sold in this studio.

    Args:
        goods (List[Good]): A list of :class:`.Good` instances to be put into this
            :class:`.PriceList` instance.
    """

    __auto_name__ = False
    __tablename__ = "PriceLists"
    __mapper_args__ = {"polymorphic_identity": "PriceList"}

    price_list_id: Mapped[int] = mapped_column(
        "id", ForeignKey("Entities.id"), primary_key=True
    )

    goods: Mapped[Optional[List["Good"]]] = relationship(
        secondary="PriceList_Goods",
        primaryjoin="PriceLists.c.id==PriceList_Goods.c.price_list_id",
        secondaryjoin="PriceList_Goods.c.good_id==Goods.c.id",
        back_populates="price_lists",
        doc="Goods in this list.",
    )

    def __init__(
        self,
        goods: Optional[List["Good"]] = None,
        **kwargs: Dict[str, Any],
    ) -> None:
        super(PriceList, self).__init__(**kwargs)
        if goods is None:
            goods = []
        self.goods = goods

    @validates("goods")
    def _validate_goods(self, key: str, good: "Good") -> "Good":
        """Validate the given good value.

        Args:
            key (str): The name of the validated column.
            good (Good): The good value to be validated.

        Raises:
            TypeError: If the given good arg value is not a :class:`.Good` instance.

        Returns:
            Good: The validated :class:`.Good` instance.
        """
        if not isinstance(good, Good):
            raise TypeError(
                f"{self.__class__.__name__}.goods should only contain "
                "instances of stalker.model.budget.Good, "
                f"not {good.__class__.__name__}: '{good}'"
            )
        return good


PriceList_Goods = Table(
    "PriceList_Goods",
    Base.metadata,
    Column("price_list_id", Integer, ForeignKey("PriceLists.id"), primary_key=True),
    Column("good_id", Integer, ForeignKey("Goods.id"), primary_key=True),
)


class Budget(Entity, ProjectMixin, DAGMixin, StatusMixin):
    """Manages project budgets.

    Budgets manager :class:`.Project` budgets. You can create entries as
    instances of :class:`.BudgetEntry` class.
    """

    __auto_name__ = False
    __tablename__ = "Budgets"
    __mapper_args__ = {"polymorphic_identity": "Budget"}

    budget_id: Mapped[int] = mapped_column(
        "id", ForeignKey("Entities.id"), primary_key=True
    )

    __id_column__ = "budget_id"

    entries: Mapped[Optional[List["BudgetEntry"]]] = relationship(
        primaryjoin="BudgetEntries.c.budget_id==Budgets.c.id",
        cascade="all, delete-orphan",
    )

    invoices: Mapped[Optional[List["Invoice"]]] = relationship(
        primaryjoin="Invoices.c.budget_id==Budgets.c.id",
        cascade="all, delete-orphan",
    )

    def __init__(self, **kwargs: Dict[str, Any]) -> None:
        super(Budget, self).__init__(**kwargs)
        ProjectMixin.__init__(self, **kwargs)
        DAGMixin.__init__(self, **kwargs)
        StatusMixin.__init__(self, **kwargs)

    @validates("entries")
    def _validate_entry(self, key: str, entry: "BudgetEntry") -> "BudgetEntry":
        """Validate the given entry value.

        Args:
            key (str): The name of the validated column.
            entry (BudgetEntry): The entry value to be validated.

        Raises:
            TypeError: If the given entry value is not a :class:`.BudgetEntry` instance.

        Returns:
            BudgetEntry: The validated BudgetEntry instance.
        """
        if not isinstance(entry, BudgetEntry):
            raise TypeError(
                f"{self.__class__.__name__}.entries should only contain "
                "instances of BudgetEntry, "
                f"not {entry.__class__.__name__}: '{entry}'"
            )
        return entry


class BudgetEntry(Entity, AmountMixin, UnitMixin):
    """Manages entries in a Budget.

    With BudgetEntries one can manage project budget entries one by one. Each
    entry shows one component of a bigger budget. Entries are generally a
    reflection of a :class:`.Good` instance and shows how many of that Good has
    been included in this Budget, and what was the discounted price of that
    Good.

    Args:
        budget (Budget): The :class:`.Budget` that this entry is a part of.
        good (Good): Stores a :class:`.Good` instance to carry all the cost/msrp/unit
            data from.
        price (float): The decided price of this entry. This is generally bigger than
            the :attr:`.cost` and should be also bigger than :attr:`.msrp` but the
            person that is editing the budget which this entry is related to can decide
            to do a discount on this entry and give a different price. This attribute
            holds the proposed final price.
        realized_total (float): This attribute is for holding the realized price of this
            entry. It can be the same number of the :attr:`.price` multiplied by the
            :attr:`.amount` or can be something else that reflects the reality.
            Generally it is for calculating the "service" cost/profit.
        amount (float): Defines the amount of :class:`Good` that is in consideration for
            this entry.
    """

    __auto_name__ = True
    __tablename__ = "BudgetEntries"
    __mapper_args__ = {"polymorphic_identity": "BudgetEntry"}

    entry_id: Mapped[int] = mapped_column(
        "id", ForeignKey("Entities.id"), primary_key=True
    )
    budget_id: Mapped[Optional[int]] = mapped_column(ForeignKey("Budgets.id"))
    budget: Mapped["Budget"] = relationship(
        primaryjoin="BudgetEntries.c.budget_id==Budgets.c.id",
        back_populates="entries",
        uselist=False,
    )

    good_id: Mapped[Optional[int]] = mapped_column(ForeignKey("Goods.id"))

    good: Mapped["Good"] = relationship(
        primaryjoin="BudgetEntries.c.good_id==Goods.c.id", uselist=False
    )

    cost: Mapped[Optional[float]] = mapped_column(default=0.0)
    msrp: Mapped[Optional[float]] = mapped_column(default=0.0)

    price: Mapped[Optional[float]] = mapped_column(default=0.0)
    realized_total: Mapped[Optional[float]] = mapped_column(default=0.0)

    def __init__(
        self,
        budget: Optional[Budget] = None,
        good: Optional[Good] = None,
        price: Union[float, int] = 0,
        realized_total: Union[float, int] = 0,
        amount: Union[float, int] = 0.0,
        **kwargs: Dict[str, Any],
    ) -> None:
        super(BudgetEntry, self).__init__(**kwargs)

        self.budget = budget
        self.good = good
        self.cost = good.cost
        self.msrp = good.msrp

        kwargs["unit"] = good.unit
        kwargs["amount"] = amount

        AmountMixin.__init__(self, **kwargs)
        UnitMixin.__init__(self, **kwargs)

        self.price = price
        self.realized_total = realized_total

    @validates("budget")
    def _validate_budget(self, key: str, budget: "Budget") -> "Budget":
        """Validate the given budget value.

        Args:
            key (str): The name of the validated column.
            budget (Budget): The budget that needs to be validated.

        Raises:
            TypeError: If the given budget is not a :class:`.Budget` instance.

        Returns:
            Budget: The validated :class:`.Budget` instance.
        """
        if not isinstance(budget, Budget):
            raise TypeError(
                f"{self.__class__.__name__}.budget should be a Budget instance, "
                f"not {budget.__class__.__name__}: '{budget}'"
            )
        return budget

    @validates("cost")
    def _validate_cost(self, key: str, cost: Union[float, int]) -> float:
        """Validate the given cost value.

        Args:
            key (str): The name of the validated column.
            cost (Union[int, float]): The cost value to be validated.

        Raises:
            TypeError: If the given cost value is not an int or float.

        Returns:
            float: The validated cost value.
        """
        if cost is None:
            cost = 0.0

        if not isinstance(cost, (int, float)):
            raise TypeError(
                f"{self.__class__.__name__}.cost should be a number, "
                f"not {cost.__class__.__name__}: '{cost}'"
            )

        return float(cost)

    @validates("msrp")
    def _validate_msrp(self, key: str, msrp: Union[float, int]) -> float:
        """Validate the given msrp value.

        Args:
            key (str): The name of the validated column.
            msrp (Union[int, float]): The msrp value to be validated.

        Raises:
            TypeError: If the given msrp value is not an int or float.

        Returns:
            float: The validated msrp value.
        """
        if msrp is None:
            msrp = 0.0

        if not isinstance(msrp, (int, float)):
            raise TypeError(
                f"{self.__class__.__name__}.msrp should be a number, "
                f"not {msrp.__class__.__name__}: '{msrp}'"
            )

        return float(msrp)

    @validates("price")
    def _validate_price(self, key: str, price: Union[float, int]) -> float:
        """Validate the given price value.

        Args:
            key (str): The name of the validated column.
            price (Union[int, float]): The price value to be validated.

        Raises:
            TypeError: If the given price value is not an int or float.

        Returns:
            float: The validated price value.
        """
        if price is None:
            price = 0.0

        if not isinstance(price, (int, float)):
            raise TypeError(
                f"{self.__class__.__name__}.price should be a number, "
                f"not {price.__class__.__name__}: '{price}'"
            )

        return float(price)

    @validates("realized_total")
    def _validate_realized_total(
        self, key: str, realized_total: Union[float, int]
    ) -> float:
        """Validate  the given realized_total value.

        Args:
            key (str): The name of the validated column.
            realized_total (Union[int, float]): The realized_total value to be
                validated.

        Raises:
            TypeError: If the given realized_total value is not an int or float.

        Returns:
            float: The validated realized_total value.
        """
        if realized_total is None:
            realized_total = 0.0

        if not isinstance(realized_total, (int, float)):
            raise TypeError(
                f"{self.__class__.__name__}.realized_total should be a number, "
                f"not {realized_total.__class__.__name__}: '{realized_total}'"
            )

        return float(realized_total)

    @validates("good")
    def _validate_good(self, key: str, good: "Good") -> "Good":
        """Validate the given good value.

        Args:
            key (str): The name of the validated column.
            good (Good): The good to be validated.

        Raises:
            TypeError: If the given good is not a :class:`.Good` instance.

        Returns:
            Good: Returns the validated :class:`.Good` instance.
        """
        if not isinstance(good, Good):
            raise TypeError(
                f"{self.__class__.__name__}.good should be a "
                "stalker.models.budget.Good instance, "
                f"not {good.__class__.__name__}: '{good}'"
            )

        return good


class Invoice(Entity, AmountMixin, UnitMixin):
    """Holds information about invoices.

    Invoices are part of :class:`.Budgets`. The main purpose of invoices are
    to track the :class:`.Payment` s. It is a very primitive entity. It is
    by no means designed to hold real financial information (at least for now).

    Args:
        client (Client): The :class:`.Client` instance that shows the payer for this
            invoice.
        budget (Budget): The :class:`.Budget` instance that owns this invoice.
        amount (Union[int, float]): The amount of this invoice. Without the
            :attr:`.Invoice.unit` attribute it is meaningless. This cannot be skipped.
        unit (str): The unit of the issued amount. This cannot be skipped.
    """

    __auto_name__ = True
    __tablename__ = "Invoices"
    __mapper_args__ = {"polymorphic_identity": "Invoice"}

    invoice_id: Mapped[int] = mapped_column(
        "id", ForeignKey("Entities.id"), primary_key=True
    )
    budget_id: Mapped[Optional[int]] = mapped_column(ForeignKey("Budgets.id"))

    budget: Mapped[Optional["Budget"]] = relationship(
        primaryjoin="Invoices.c.budget_id==Budgets.c.id",
        back_populates="invoices",
        uselist=False,
    )

    client_id: Mapped[Optional[int]] = mapped_column(ForeignKey("Clients.id"))

    client: Mapped[Optional["Client"]] = relationship(
        primaryjoin="Invoices.c.client_id==Clients.c.id", uselist=False
    )

    payments: Mapped[Optional[List["Payment"]]] = relationship(
        primaryjoin="Payments.c.invoice_id==Invoices.c.id",
        cascade="all, delete-orphan",
    )

    def __init__(
        self,
        budget: Optional["Budget"] = None,
        client: Optional["Client"] = None,
        amount: Union[float, int] = 0,
        unit: Optional[str] = None,
        **kwargs,
    ) -> None:
        super(Invoice, self).__init__(**kwargs)
        AmountMixin.__init__(self, amount=amount)
        UnitMixin.__init__(self, unit=unit)
        self.budget = budget
        self.client = client

    @validates("budget")
    def _validate_budget(self, key: str, budget: "Budget") -> "Budget":
        """Validate the given budget value.

        Args:
            key (str): The name of the validated column.
            budget (Budget): The :class:`.Budget` instance to be validated.

        Raises:
            TypeError: If the given budget arg value is not a :class:`.Budget` instance.

        Returns:
            Budget: The validated :class:`.Budget` instance.
        """
        if not isinstance(budget, Budget):
            raise TypeError(
                f"{self.__class__.__name__}.budget should be a Budget instance, "
                f"not {budget.__class__.__name__}: '{budget}'"
            )
        return budget

    @validates("client")
    def _validate_client(self, key: str, client: "Client") -> "Client":
        """Validate the given client value.

        Args:
            key (str): The name of the validated column.
            client (Client): The :class:`stalker.models.client.Client` instance to be
                validated.

        Raises:
            TypeError: If the ``client`` is not a :class:`stalker.models.client.Client`
                instance.

        Returns:
            Client: The validated :class:`stalker.models.client.Client` instance.
        """
        from stalker.models.client import Client

        if not isinstance(client, Client):
            raise TypeError(
                f"{self.__class__.__name__}.client should be a Client instance, "
                f"not {client.__class__.__name__}: '{client}'"
            )
        return client


class Payment(Entity, AmountMixin, UnitMixin):
    """Holds information about the payments.

    Each payment should be related with an :class:`.Invoice` instance. Use the
    :attr:`.type` attribute to diversify payments (ex. "Advance").

    Args:
        invoice (Invoice): The :class:`.Invoice` instance that this payment is related
            to. This cannot be skipped.
        amount (Union[int, float]): The amount value.
        unit (Optional[str]): The unit of this mixed in class.
    """

    __auto_name__ = True
    __tablename__ = "Payments"
    __mapper_args__ = {"polymorphic_identity": "Payment"}

    payment_id: Mapped[int] = mapped_column(
        "id", ForeignKey("Entities.id"), primary_key=True
    )
    invoice_id: Mapped[Optional[int]] = mapped_column(ForeignKey("Invoices.id"))
    invoice: Mapped[Optional["Invoice"]] = relationship(
        primaryjoin="Payments.c.invoice_id==Invoices.c.id",
        back_populates="payments",
        uselist=False,
    )

    def __init__(
        self,
        invoice: Optional["Invoice"] = None,
        amount: Union[int, float] = 0,
        unit: Optional[str] = None,
        **kwargs,
    ) -> None:
        super(Payment, self).__init__(**kwargs)
        AmountMixin.__init__(self, amount=amount)
        UnitMixin.__init__(self, unit=unit)
        self.invoice = invoice

    @validates("invoice")
    def _validate_invoice(self, key: str, invoice: "Invoice") -> "Invoice":
        """Validate the invoice value.

        Args:
            key (str): The name of the validated column.
            invoice (Invoice): The :class:`.Invoice` instance to validate.

        Raises:
            TypeError: The :class:`.Invoice` instance to be validated.

        Returns:
            Invoice: The validated :class:`.Invoice` instance.
        """
        if not isinstance(invoice, Invoice):
            raise TypeError(
                f"{self.__class__.__name__}.invoice should be an Invoice instance, "
                f"not {invoice.__class__.__name__}: '{invoice}'"
            )
        return invoice



# -*- coding: utf-8 -*-
"""Type related functions and classes are situated here."""

from typing import Any, Dict, Optional

from sqlalchemy import ForeignKey, String
from sqlalchemy.orm import Mapped, mapped_column

from stalker.db.declarative import Base
from stalker.log import get_logger
from stalker.models.entity import Entity
from stalker.models.mixins import CodeMixin, TargetEntityTypeMixin

logger = get_logger(__name__)


class Type(Entity, TargetEntityTypeMixin, CodeMixin):
    """Everything can have a type.

    .. versionadded:: 0.1.1
      Types

    Type is a generalized version of the previous design that defines types for
    specific classes.

    The purpose of the :class:`.Type` class is just to define a new type for a
    specific :class:`.Entity`. For example, you can have a ``Character``
    :class:`.Asset` or you can have a ``Commercial`` :class:`.Project` or you
    can define a :class:`.File` as an ``Image`` etc., to create a new
    :class:`.Type` for various classes:

    ..code-block: Python

        Type(name="Character", target_entity_type="Asset")
        Type(name="Commercial", target_entity_type="Project")
        Type(name="Image", target_entity_type="File")

    or:

    ..code-block: Python

        Type(name="Character", target_entity_type=Asset.entity_type)
        Type(name="Commercial", target_entity_type=Project.entity_type)
        Type(name="Image", target_entity_type=File.entity_type)

    or even better:

    ..code-block: Python

        Type(name="Character", target_entity_type=Asset)
        Type(name="Commercial", target_entity_type=Project)
        Type(name="Image", target_entity_type=File)

    By using :class:`.Type` s, one can able to sort and group same type of
    entities.

    :class:`.Type` s are generally used in :class:`.Structure` s.

    Args:
        target_entity_type (str): The string defining the target type of this
            :class:`.Type`.
    """

    __auto_name__ = False
    __tablename__ = "Types"
    __mapper_args__ = {"polymorphic_identity": "Type"}
    type_id_local: Mapped[int] = mapped_column(
        "id",
        ForeignKey("Entities.id"),
        primary_key=True,
    )

    def __init__(
        self,
        name: Optional[str] = None,
        code: Optional[str] = None,
        target_entity_type: Optional[str] = None,
        **kwargs: Dict[str, Any],
    ) -> None:
        kwargs["name"] = name
        kwargs["target_entity_type"] = target_entity_type
        super(Type, self).__init__(**kwargs)
        TargetEntityTypeMixin.__init__(self, **kwargs)
        # CodeMixin.__init__(self, **kwargs)
        self.code = code

    def __eq__(self, other: Any) -> bool:
        """Check the equality.

        Args:
            other (Any): The other object.

        Returns:
            bool: True if the other object is equal to this Type instance as an Entity
                and has the same target_entity_type.
        """
        return (
            super(Type, self).__eq__(other)
            and isinstance(other, Type)
            and self.target_entity_type == other.target_entity_type
        )

    def __hash__(self) -> int:
        """Return the hash value of this instance.

        Because the __eq__ is overridden the __hash__ also needs to be overridden.

        Returns:
            int: The hash value.
        """
        return super(Type, self).__hash__()


class EntityType(Base):
    """A simple class just to hold the registered class names in Stalker."""

    __tablename__ = "EntityTypes"
    __table_args__ = {"extend_existing": True}

    id: Mapped[int] = mapped_column("id", primary_key=True)
    name: Mapped[str] = mapped_column(
        String(128),
        nullable=False,
        unique=True,
    )
    statusable: Mapped[Optional[bool]] = mapped_column(default=False)
    dateable: Mapped[Optional[bool]] = mapped_column(default=False)
    schedulable: Mapped[Optional[bool]] = mapped_column(default=False)
    accepts_references: Mapped[Optional[bool]] = mapped_column(default=False)

    def __init__(
        self,
        name: str,
        statusable: bool = False,
        schedulable: bool = False,
        accepts_references: bool = False,
    ) -> None:
        self.name = name
        self.statusable = statusable
        self.schedulable = schedulable
        self.accepts_references = accepts_references

        # TODO: add tests for the name attribute



# -*- coding: utf-8 -*-
"""Stalker is a Production Asset Management (ProdAM) designed for Animation/VFX Studios.

See docs for more information.
"""
from stalker.version import __version__  # noqa: F401
from stalker import config, log  # noqa: I100

if True:
    defaults: config.Config = config.Config()
from stalker.models.asset import Asset
from stalker.models.auth import (
    AuthenticationLog,
    Group,
    LocalSession,
    Permission,
    Role,
    User,
)
from stalker.models.budget import Budget, BudgetEntry, Good, Invoice, Payment, PriceList
from stalker.models.client import Client, ClientUser
from stalker.models.department import Department, DepartmentUser
from stalker.models.entity import Entity, EntityGroup, SimpleEntity
from stalker.models.format import ImageFormat
from stalker.models.file import File
from stalker.models.message import Message
from stalker.models.mixins import (
    ACLMixin,
    AmountMixin,
    CodeMixin,
    DAGMixin,
    DateRangeMixin,
    ProjectMixin,
    ReferenceMixin,
    ScheduleMixin,
    StatusMixin,
    TargetEntityTypeMixin,
    UnitMixin,
    WorkingHoursMixin,
)
from stalker.models.note import Note
from stalker.models.project import (
    Project,
    ProjectClient,
    ProjectRepository,
    ProjectUser,
)
from stalker.models.repository import Repository
from stalker.models.review import Daily, DailyFile, Review
from stalker.models.scene import Scene
from stalker.models.schedulers import SchedulerBase, TaskJugglerScheduler
from stalker.models.sequence import Sequence
from stalker.models.shot import Shot
from stalker.models.status import Status, StatusList
from stalker.models.structure import Structure
from stalker.models.studio import Studio, Vacation, WorkingHours
from stalker.models.tag import Tag
from stalker.models.task import Task, TaskDependency, TimeLog
from stalker.models.template import FilenameTemplate
from stalker.models.ticket import Ticket, TicketLog
from stalker.models.type import EntityType, Type
from stalker.models.variant import Variant
from stalker.models.version import Version
from stalker.models.wiki import Page

__all__ = [
    "ACLMixin",
    "AmountMixin",
    "Asset",
    "AuthenticationLog",
    "Budget",
    "BudgetEntry",
    "Client",
    "ClientUser",
    "CodeMixin",
    "DAGMixin",
    "Daily",
    "DailyFile",
    "DateRangeMixin",
    "Department",
    "DepartmentUser",
    "Entity",
    "EntityGroup",
    "EntityType",
    "File",
    "FilenameTemplate",
    "Good",
    "Group",
    "ImageFormat",
    "Invoice",
    "LocalSession",
    "Message",
    "Note",
    "Page",
    "Payment",
    "Permission",
    "PriceList",
    "Project",
    "ProjectClient",
    "ProjectMixin",
    "ProjectRepository",
    "ProjectUser",
    "ReferenceMixin",
    "Repository",
    "Review",
    "Role",
    "Scene",
    "ScheduleMixin",
    "SchedulerBase",
    "Sequence",
    "Shot",
    "SimpleEntity",
    "Status",
    "StatusList",
    "StatusMixin",
    "Structure",
    "Studio",
    "Tag",
    "TargetEntityTypeMixin",
    "Task",
    "TaskDependency",
    "TaskJugglerScheduler",
    "Ticket",
    "TicketLog",
    "TimeLog",
    "Type",
    "UnitMixin",
    "User",
    "Vacation",
    "Variant",
    "Version",
    "WorkingHours",
    "WorkingHoursMixin",
]


logger = log.get_logger(__name__)



# -*- coding: utf-8 -*-
"""Scene related classes and functions are situated here."""

from typing import Any, Dict, List, Optional, TYPE_CHECKING

from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates

from stalker.log import get_logger
from stalker.models.mixins import CodeMixin
from stalker.models.task import Task

if TYPE_CHECKING:  # pragma: no cover
    from stalker.models.shot import Shot

logger = get_logger(__name__)


class Scene(Task, CodeMixin):
    """Stores data about Scenes.

    Scenes are grouping the Shots according to their view to the world, that is
    shots taking place in the same set configuration can be grouped together by
    using Scenes.

    You cannot replace :class:`.Sequence` s with Scenes, because Scene
    instances doesn't have some key features that :class:`.Sequence` s have.

    A Scene needs to be tied to a :class:`.Project`
    instance, so it is not possible to create a Scene without a one.
    """

    __auto_name__ = False
    __tablename__ = "Scenes"
    __mapper_args__ = {"polymorphic_identity": "Scene"}
    scene_id: Mapped[int] = mapped_column(
        "id",
        ForeignKey("Tasks.id"),
        primary_key=True,
    )

    shots: Mapped[Optional[List["Shot"]]] = relationship(
        primaryjoin="Shots.c.scene_id==Scenes.c.id",
        back_populates="scene",
        doc="""The :class:`.Shot` s that is related with this Scene.

        It is a list of :class:`.Shot` instances.
        """,
    )

    def __init__(self, shots: Optional[List["Shot"]] = None, **kwargs: Dict[str, Any]):
        super(Scene, self).__init__(**kwargs)

        # call the mixin __init__ methods
        CodeMixin.__init__(self, **kwargs)

        if shots is None:
            shots = []

        self.shots = shots

    @validates("shots")
    def _validate_shots(self, key: str, shot: "Shot") -> "Shot":
        """Validate the given shot value.

        Args:
            key (str): The name of the validated column.
            shot (Shot): The shot instance.

        Raises:
            TypeError: If the shot is not a Shot instance.

        Returns:
            Shot: Return the validated Shot instance.
        """
        from stalker.models.shot import Shot

        if not isinstance(shot, Shot):
            raise TypeError(
                f"{self.__class__.__name__}.shots should only contain "
                "instances of stalker.models.shot.Shot, "
                f"not {shot.__class__.__name__}: '{shot}'"
            )
        return shot

    def __eq__(self, other: Any) -> bool:
        """Check the equality with the other object.

        Args:
            other (Any): The other object.

        Returns:
            bool: True if the other object is equal to this object.
        """
        return isinstance(other, Scene) and super(Scene, self).__eq__(other)

    def __hash__(self) -> int:
        """Return the hash value of this instance.

        Because the __eq__ is overridden the __hash__ also needs to be overridden.

        Returns:
            int: The hash value.
        """
        return super(Scene, self).__hash__()



# -*- coding: utf-8 -*-
"""Note class lies here."""
from typing import Any, Dict, Optional

from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, synonym

from stalker.log import get_logger
from stalker.models.entity import SimpleEntity

logger = get_logger(__name__)


class Note(SimpleEntity):
    """Notes for any of the SOM objects.

    To leave notes in Stalker use the Note class.

    Args:
        content (str): The content of the note.
        attached_to (Entity): The object that this note is attached to.
    """

    __auto_name__ = True
    __tablename__ = "Notes"
    __mapper_args__ = {"polymorphic_identity": "Note"}

    note_id: Mapped[int] = mapped_column(
        "id",
        ForeignKey("SimpleEntities.id"),
        primary_key=True,
    )

    content: Mapped[Optional[str]] = synonym(
        "description",
        doc="""The content of this :class:`.Note` instance.

        Content is a string representing the content of this Note, can be an
        empty.
        """,
    )

    def __init__(self, content: str = "", **kwargs: Dict[str, Any]) -> None:
        super(Note, self).__init__(**kwargs)
        self.content = content

    def __eq__(self, other: Any) -> bool:
        """Check the equality.

        Args:
            other (Any): The other object.

        Returns:
            bool: True if the other object is a Note instance and has the same content.
        """
        return (
            super(Note, self).__eq__(other)
            and isinstance(other, Note)
            and self.content == other.content
        )

    def __hash__(self) -> int:
        """Return the hash value of this instance.

        Because the __eq__ is overridden the __hash__ also needs to be overridden.

        Returns:
            int: The hash value.
        """
        return super(Note, self).__hash__()



# -*- coding: utf-8 -*-
"""Repository related functionality is situated here."""
import os
import platform
from typing import Any, Dict, Optional, TYPE_CHECKING

from sqlalchemy import ForeignKey, String, event
from sqlalchemy.orm import Mapped, mapped_column, validates

from stalker import defaults
from stalker.log import get_logger
from stalker.models.entity import Entity
from stalker.models.mixins import CodeMixin

if TYPE_CHECKING:  # pragma: no cover
    from sqlalchemy.orm import Mapper
    from sqlalchemy.engine import Connection

logger = get_logger(__name__)


class Repository(Entity, CodeMixin):
    r"""Manage fileserver/repository related data.

    A repository is a network share that all users have access to.

    A studio can create several repositories, for example, one for movie
    projects and one for commercial projects.

    A repository also defines the default paths for linux, windows and mac
    foreshores.

    The path separator in the repository is always forward slashes ("/").
    Setting a path that contains backward slashes ("\"), will be converted to
    a path with forward slashes.

    .. versionadded:: 0.2.24
       Code attribute

       Starting with v0.2.24 Repository instances have a new :attr:`.code`
       attribute whose value is used by the
       :class:`stalker.models.studio.Studio` to generate environment variables
       that contains the path of this
       :class:`stalker.models.repository.Repository` (i.e.
       $REPOCP/path/to/asset.ma ``CP`` here is the ``Repository.code``) so that
       instead of using absolute full paths one can use the
       :attr:`.make_relative`` path to generate a universal path that can be
       used across OSes and different installations of Stalker.

    Args:
        code (str): The code of the :class:`stalker.models.repository.Repository`.
            This attribute value is used by the :class:`stalker.models.studio.Studio`
            to generate environment variables that contains the path of this
            ``Repository`` (i.e. $REPOCP/path/to/asset.ma) so that instead of
            using absolute full paths one can use the ``repository_relative``
            path to generate a universal path that can be used across OSes and
            different installations of Stalker.
        linux_path (str): shows the linux path of the repository root, should be
            a string
        macos_path (str): shows the macOS path of the repository root, should be
            a string.
        windows_path (str): shows the windows path of the repository root, should
            be a string
    """

    #
    # TODO: Add OpenLDAP support.
    #
    # In an OpenLDAP Server + AutoFS setup Stalker can create new entries to
    # OpenLDAP server.
    #
    # The AutoFS can be installed to any linux system easily or it is already
    # installed. macOS has it already. I know nothing about Windows.
    #
    # AutoFS can be setup to listen for new mount points from an OpenLDAP
    # server. Thus it is heavily related with the users system, Stalker
    # cannot do anything about that. The IT should setup workstations.
    #
    # But Stalker can connect to the OpenLDAP server and create new entries.
    #

    __auto_name__ = False
    __tablename__ = "Repositories"
    __mapper_args__ = {"polymorphic_identity": "Repository"}
    repository_id: Mapped[int] = mapped_column(
        "id",
        ForeignKey("Entities.id"),
        primary_key=True,
    )
    linux_path: Mapped[Optional[str]] = mapped_column(String(256))
    windows_path: Mapped[Optional[str]] = mapped_column(String(256))
    macos_path: Mapped[Optional[str]] = mapped_column(String(256))

    def __init__(
        self,
        code: str = "",
        linux_path: str = "",
        windows_path: str = "",
        macos_path: str = "",
        **kwargs: Dict[str, Any],
    ) -> None:
        kwargs["code"] = code
        super(Repository, self).__init__(**kwargs)
        CodeMixin.__init__(self, **kwargs)

        self.linux_path = linux_path
        self.windows_path = windows_path
        self.macos_path = macos_path

    @validates("linux_path")
    def _validate_linux_path(self, key: str, linux_path: str) -> str:
        """Validate the given Linux path.

        Args:
            key (str): The name of the validated column.
            linux_path (str): The Linux path to validated.

        Raises:
            TypeError: If the given Linux path is not a str.

        Returns:
            str: The validated Linux path.
        """
        if not isinstance(linux_path, str):
            raise TypeError(
                f"{self.__class__.__name__}.linux_path should be an instance of "
                f"string, not {linux_path.__class__.__name__}: '{linux_path}'"
            )

        linux_path = os.path.normpath(linux_path) + "/"

        linux_path = linux_path.replace("\\", "/")

        if self.code is not None and platform.system() == "Linux":
            # update the environment variable
            os.environ[defaults.repo_env_var_template.format(code=self.code)] = (
                linux_path
            )

        return linux_path

    @validates("macos_path")
    def _validate_macos_path(self, key: str, macos_path: str) -> str:
        """Validate the given macOS path.

        Args:
            key (str): The name of the validated column.
            macos_path (str): The macOS path to validate.

        Raises:
            TypeError: If the given macOS path is not a str.

        Returns:
            str: The validated macOS path.
        """
        if not isinstance(macos_path, str):
            raise TypeError(
                f"{self.__class__.__name__}.macos_path should be an instance of "
                f"string, not {macos_path.__class__.__name__}: '{macos_path}'"
            )

        macos_path = os.path.normpath(macos_path) + "/"
        macos_path = macos_path.replace("\\", "/")
        if self.code is not None and platform.system() == "Darwin":
            # update the environment variable
            rendered_env_var = defaults.repo_env_var_template.format(code=self.code)
            os.environ[rendered_env_var] = macos_path

        return macos_path

    @validates("windows_path")
    def _validate_windows_path(self, key: str, windows_path: str) -> str:
        """Validate the given Windows path.

        Args:
            key (str): The name of the validated column.
            windows_path (str): The Windows path to validate.

        Raises:
            TypeError: If the given Windows path is not a str.

        Returns:
            str: The validated Windows path.
        """
        if not isinstance(windows_path, str):
            raise TypeError(
                f"{self.__class__.__name__}.windows_path should be an instance of "
                f"string, not {windows_path.__class__.__name__}: '{windows_path}'"
            )

        windows_path = os.path.normpath(windows_path)
        windows_path = windows_path.replace("\\", "/")

        if not windows_path.endswith("/"):
            windows_path += "/"

        if self.code is not None and platform.system() == "Windows":
            # update the environment variable
            os.environ[defaults.repo_env_var_template.format(code=self.code)] = (
                windows_path
            )

        return windows_path

    @property
    def path(self) -> str:
        """Return the repository path for the current OS.

        Returns:
            str: The repository path for the current OS.
        """
        # return the proper value according to the current os
        platform_system = platform.system()

        if platform_system == "Linux":
            return self.linux_path
        elif platform_system == "Windows":
            return self.windows_path
        elif platform_system == "Darwin":
            return self.macos_path

    @path.setter
    def path(self, path: str) -> None:
        """Set the path for the current OS.

        Args:
            path (str): The path.
        """
        # return the proper value according to the current os
        platform_system = platform.system()

        if platform_system == "Linux":
            self.linux_path = path
        elif platform_system == "Windows":
            self.windows_path = path
        elif platform_system == "Darwin":
            self.macos_path = path

    def is_in_repo(self, path: str) -> bool:
        """Return True or False depending on the given is in this repo or not.

        Args:
            path: The path to be investigated.

        Returns:
            bool: Return True if the given path is in this repository.
        """
        path = path.replace("\\", "/")
        return (
            path.lower().startswith(self.windows_path.lower())
            or path.startswith(self.linux_path)
            or path.startswith(self.macos_path)
        )

    def _to_path(self, path: str, replace_with: str) -> str:
        """Return the path replacing the OS related part with the given str.

        Args:
            path (str): The input path.
            replace_with (str): replace_with path

        Raises:
            TypeError: When the given path is not a str.

        Returns:
            str: The converted path.
        """
        if not isinstance(path, str):
            raise TypeError(
                "path should be a string containing a file path, "
                f"not {path.__class__.__name__}: '{path}'"
            )

        if not isinstance(replace_with, str):
            raise TypeError(
                "replace_with should be a string containing a file path, "
                f"not {replace_with.__class__.__name__}: '{replace_with}'"
            )

        # expand all variables
        path = os.path.normpath(os.path.expandvars(os.path.expanduser(path))).replace(
            "\\", "/"
        )

        if path.startswith(self.windows_path):
            return path.replace(self.windows_path, replace_with)
        elif path.startswith(self.linux_path):
            return path.replace(self.linux_path, replace_with)
        elif path.startswith(self.macos_path):
            return path.replace(self.macos_path, replace_with)

        return path

    def to_linux_path(self, path: str) -> str:
        """Return the Linux version of the given path.

        Args:
            path (str): The path that needs to be converted to Linux path.

        Returns:
            str: The Linux path.
        """
        return self._to_path(path, self.linux_path)

    def to_windows_path(self, path: str) -> str:
        """Return the Windows version of the given path.

        Args:
            path (str): The path that needs to be converted to windows path.

        Returns:
            str: The Windows path.
        """
        return self._to_path(path, self.windows_path)

    def to_macos_path(self, path: str) -> str:
        """Return the macOS version of the given path.

        Args:
            path (str): The path that needs to be converted to macOS path.

        Returns:
            str: The macOS path.
        """
        return self._to_path(path, self.macos_path)

    def to_native_path(self, path: str) -> str:
        """Return the native version of the given path.

        Args:
            path (str): The path that needs to be converted to native path.

        Returns:
            str: The native path.
        """
        return self._to_path(path, self.path)

    def make_relative(self, path: str) -> str:
        """Make the given path relative to the repository root.

        Args:
            path (str): The path to be made relative.

        Returns:
            str: The relative path.
        """
        path = self.to_native_path(path)
        return os.path.relpath(path, self.path).replace("\\", "/")

    @classmethod
    def find_repo(cls, path: str) -> "Repository":
        """Return the repository from the given path.

        Args:
            path (str): Path in a repository.

        Returns:
            Repository:
        """
        logger.debug(f"Looking for a repo for path: {path}")
        # path could be using environment variables so expand them
        path = os.path.expandvars(path)
        logger.debug(f"path after expanding vars  : {path}")

        # first find the repository
        repos = Repository.query.all()
        found_repo = None
        for repo in repos:
            if (
                path.startswith(repo.path)
                or path.lower().startswith(repo.windows_path.lower())
                or path.startswith(repo.linux_path)
                or path.startswith(repo.macos_path)
            ):
                found_repo = repo
                break

        if found_repo is None:
            logger.debug(f"Couldn't find a repo for path: {path}")

        return found_repo

    @classmethod
    def to_os_independent_path(cls, path: str) -> str:
        """Replace the part of the given path with repository environment var.

         This makes the given path OS independent.

        Args:
            path (str): path to make OS independent.

        Returns:
            str: OS independent path.
        """
        # find the related repo
        repo = cls.find_repo(path)

        if repo:
            logger.debug("Found repo for path: {}".format(repo))
            return "${}/{}".format(repo.env_var, repo.make_relative(path))
        else:
            logger.debug("Can't find repo for path: {}".format(path))
            return path

    @property
    def env_var(self) -> str:
        """Return the env var of this repo.

        Returns:
            str: The env_var corresponding to this repo.
        """
        return defaults.repo_env_var_template.format(code=self.code)

    def __eq__(self, other: Any) -> bool:
        """Check the equality.

        Args:
            other (Any): The other object.

        Returns:
            bool: True if the other object is equal to this one as an Entity, is a
                Repository instance and has the same linux_path, macos_path,
                windows_path.
        """
        return (
            super(Repository, self).__eq__(other)
            and isinstance(other, Repository)
            and self.linux_path == other.linux_path
            and self.macos_path == other.macos_path
            and self.windows_path == other.windows_path
        )

    def __hash__(self) -> int:
        """Return the hash value of this instance.

        Because the __eq__ is overridden the __hash__ also needs to be overridden.

        Returns:
            int: The hash value.
        """
        return super(Repository, self).__hash__()


@event.listens_for(Repository, "after_insert")
def receive_after_insert(
    mapper: "Mapper",
    connection: "Connection",
    repo: "Repository",
) -> None:
    """Listen for the 'after_insert' event and update environment variables.

    This is a mapper event to update the environment variables with the newly inserted
    Repository data.

    Args:
        mapper (sqlalchemy.orm.Mapper): The mapper object.
        connection (sqlalchemy.engine.Connection): The connection object.
        repo (Repository): The Repository instance that is just inserted to the DB.
    """
    logger.debug("auto creating env var for Repository: {}".format(repo.name))
    os.environ[defaults.repo_env_var_template.format(code=repo.code)] = repo.path



# -*- coding: utf-8 -*-
"""Studio, WorkingHours and Vacation related functions and classes are situated here."""

import copy
import datetime
import time
from math import ceil
from typing import Any, Dict, List, Optional, Union

import pytz

from sqlalchemy import ForeignKey, Interval, Text
from sqlalchemy.orm import (
    Mapped,
    mapped_column,
    reconstructor,
    relationship,
    synonym,
    validates,
)

from stalker import defaults, log
from stalker.db.session import DBSession
from stalker.db.types import GenericDateTime, GenericJSON
from stalker.models.auth import User
from stalker.models.department import Department
from stalker.models.entity import Entity, SimpleEntity
from stalker.models.mixins import DateRangeMixin, WorkingHoursMixin
from stalker.models.project import Project
from stalker.models.schedulers import SchedulerBase
from stalker.models.status import Status


logger = log.get_logger(__name__)


class Studio(Entity, DateRangeMixin, WorkingHoursMixin):
    """Manage all the studio information at once.

    With Stalker, you can manage all you Studio data by using this class. Studio
    knows all the projects, all the departments, all the users and every thing
    about the studio. But the most important part of the Studio is that it can
    schedule all the Projects by using TaskJuggler.

    Studio class is kind of the database itself::

      studio = Studio()

      # simple data
      studio.projects
      studio.active_projects
      studio.inactive_projects
      studio.departments
      studio.users

      # project management
      studio.to_tjp          # a tjp representation of the studio with all
                             # its projects, departments and resources etc.

      studio.schedule() # schedules all the active projects at once

    **Working Hours**

    In Stalker, Studio class also manages the working hours of the studio.
    Allowing project tasks to be scheduled to be scheduled in those hours.

    **Vacations**

    Studio wide vacations are managed by the Studio class.

    **Scheduling**

    .. versionadded: 0.2.5
       Schedule Info Attributes

    There are a couple of attributes those become pretty interesting when used
    together with the Studio instance while using the scheduling part of the
    Studio. Please refer to the attribute documentation for each attribute:

      :attr:`.is_scheduling`
      :attr:`.last_scheduled_at`
      :attr:`.last_scheduled_by`
      :attr:`.last_schedule_message`

    Args:
        daily_working_hours (int): An integer specifying the daily working
            hours for the studio. It is another critical value attribute which
            TaskJuggler uses mainly converting working day values to working hours
            (1d = 10h kind of thing).
        now (datetime.datetime): The now attribute overrides the TaskJugglers ``now``
            attribute allowing the user to schedule the projects as if the scheduling is
            done on that date. The default value is the rounded value of
            datetime.datetime.now(pytz.utc).
        timing_resolution (datetime.timedelta): The timing_resolution of the
            datetime.datetime object in datetime.timedelta. Uses ``timing_resolution``
            settings in the :class:`stalker.config.Config` class which defaults to 1
            hour. Setting the timing_resolution to less then 5 minutes is not suggested
            because it is a limit for TaskJuggler.
    """

    __auto_name__ = False
    __tablename__ = "Studios"
    __mapper_args__ = {"polymorphic_identity": "Studio"}

    studio_id: Mapped[int] = mapped_column(
        "id",
        ForeignKey("Entities.id"),
        primary_key=True,
    )

    _timing_resolution: Mapped[Optional[datetime.timedelta]] = mapped_column(
        "timing_resolution", Interval
    )

    is_scheduling: Mapped[Optional[bool]] = mapped_column(default=False)
    is_scheduling_by_id: Mapped[Optional[int]] = mapped_column(
        ForeignKey("Users.id"),
        doc="The id of the user who is scheduling the Studio projects right " "now",
    )
    is_scheduling_by: Mapped[Optional[User]] = relationship(
        primaryjoin="Studios.c.is_scheduling_by_id==Users.c.id",
        doc="The User who is scheduling the Studio projects right now",
    )
    scheduling_started_at: Mapped[Optional[datetime.datetime]] = mapped_column(
        GenericDateTime,
        doc="Stores when the current scheduling is started at, it is a good "
        "measure for measuring if the last schedule is not correctly "
        "finished",
    )
    last_scheduled_at: Mapped[Optional[datetime.datetime]] = mapped_column(
        GenericDateTime, doc="Stores the last schedule date"
    )
    last_scheduled_by_id: Mapped[Optional[int]] = mapped_column(
        ForeignKey("Users.id"),
        doc="The id of the user who has last scheduled the Studio projects",
    )
    last_scheduled_by: Mapped[Optional[User]] = relationship(
        primaryjoin="Studios.c.last_scheduled_by_id==Users.c.id",
        doc="The User who has last scheduled the Studio projects",
    )
    last_schedule_message: Mapped[Optional[str]] = mapped_column(
        Text,
        doc="Holds the last schedule message, generally coming generated by "
        "TaskJuggler",
    )

    def __init__(
        self,
        daily_working_hours: Optional[int] = None,
        now: Optional[datetime.datetime] = None,
        timing_resolution: Optional[datetime.timedelta] = None,
        **kwargs: Dict[str, Any],
    ) -> None:
        super(Studio, self).__init__(**kwargs)
        DateRangeMixin.__init__(self, **kwargs)
        WorkingHoursMixin.__init__(self, **kwargs)
        self.timing_resolution = timing_resolution
        self.daily_working_hours = daily_working_hours
        self._now = None
        self.now = self._validate_now(now)
        self._scheduler = None

        # update defaults
        self.update_defaults()

    @property
    def daily_working_hours(self) -> int:
        """Return the Studio.working_hours.daily_working_hours.

        Returns:
            int: The daily working hours of this Studio.
        """
        return self.working_hours.daily_working_hours

    @daily_working_hours.setter
    def daily_working_hours(self, daily_working_hours: int) -> None:
        """Set the Studio.working_hours.daily_working_hours.

        Args:
            daily_working_hours (int): The daily working hours in this studio.
        """
        self.working_hours.daily_working_hours = daily_working_hours

    def update_defaults(self) -> None:
        """Update the default values with the studio."""
        # TODO: add update_defaults() to attribute edit/update methods,
        #       so we will always have an up to date info about the working
        #       hours.

        logger.debug("updating defaults with Studio instance")
        logger.debug("defaults: {}".format(defaults))
        logger.debug("id(defaults): {}".format(id(defaults)))

        defaults["daily_working_hours"] = self.daily_working_hours
        logger.debug(
            "updated defaults.daily_working_hours: {}".format(
                defaults.daily_working_hours
            )
        )

        defaults["weekly_working_days"] = self.weekly_working_days
        logger.debug(
            f"updated defaults.weekly_working_days: {defaults.weekly_working_days}"
        )

        defaults["weekly_working_hours"] = self.weekly_working_hours
        logger.debug(
            "updated defaults.weekly_working_hours: {}".format(
                defaults.weekly_working_hours
            )
        )

        defaults["yearly_working_days"] = self.yearly_working_days
        logger.debug(
            f"updated defaults.yearly_working_days: {defaults.yearly_working_days}"
        )

        defaults["timing_resolution"] = self.timing_resolution
        logger.debug(
            f"updated defaults.timing_resolution: {defaults.timing_resolution}"
        )

        logger.debug(
            """done updating defaults:
        daily_working_hours  : {daily_working_hours}
        weekly_working_days  : {weekly_working_days}
        weekly_working_hours : {weekly_working_hours}
        yearly_working_days  : {yearly_working_days}
        timing_resolution    : {timing_resolution}
        """.format(
                daily_working_hours=defaults.daily_working_hours,
                weekly_working_days=defaults.weekly_working_days,
                weekly_working_hours=defaults.weekly_working_hours,
                yearly_working_days=defaults.yearly_working_days,
                timing_resolution=defaults.timing_resolution,
            )
        )

    @reconstructor
    def __init_on_load__(self) -> None:
        """Update defaults on load."""
        self.update_defaults()

    def _validate_now(self, now: datetime.datetime) -> datetime.datetime:
        """Validate the given now value.

        Args:
            now (Union[None, datetime.datetime]): Either None in which the current
                date and time will be used or a datetime.datetime instance.

        Raises:
            TypeError: If the now value is not None and not a datetime.datetime
                instance.

        Returns:
            datetime.datetime: The validated datetime.datetime value.
        """
        if now is None:
            now = datetime.datetime.now(pytz.utc)

        if not isinstance(now, datetime.datetime):
            raise TypeError(
                "{}.now attribute should be an instance of datetime.datetime, "
                "not {}: '{}'".format(
                    self.__class__.__name__, now.__class__.__name__, now
                )
            )

        return self.round_time(now)

    @property
    def now(self) -> datetime.datetime:
        """Return the currently stored now value.

        Returns:
            datetime.datetime: Return the currently stored now value if there is any,
                return the current date and time otherwise.
        """
        if self._now is None:
            self._now = self.round_time(datetime.datetime.now(pytz.utc))
        return self._now

    @now.setter
    def now(self, now: datetime.datetime) -> None:
        """Set the current date and time.

        Args:
            now (datetime.datetime): The datetime.datetime instance showing the current
                date and time, useful for project management purposes before
                scheduling.
        """
        self._now = self._validate_now(now)

    def _validate_scheduler(
        self, scheduler: Union[None, SchedulerBase]
    ) -> Union[None, SchedulerBase]:
        """Validate the given scheduler value.

        Args:
            scheduler (Union[None, SchedulerBase]): The scheduler to be used to schedule
                the projects in this Studio instance. Can be set to None to disable the
                scheduling abilities.

        Raises:
            TypeError: If the given scheduler value is not None and is not a
                SchedulerBase instance.

        Returns:
            Union[None, SchedulerBase]: The validated scheduler value.
        """
        if scheduler is not None and not isinstance(scheduler, SchedulerBase):
            raise TypeError(
                "{}.scheduler should be an instance of "
                "stalker.models.scheduler.SchedulerBase, not {}: '{}'".format(
                    self.__class__.__name__, scheduler.__class__.__name__, scheduler
                )
            )
        return scheduler

    @property
    def scheduler(self) -> Union[None, SchedulerBase]:
        """Return the scheduler.

        Returns:
            Union[None, SchedulerBase]: The scheduler of this Studio.
        """
        return self._scheduler

    @scheduler.setter
    def scheduler(self, scheduler: Union[None, SchedulerBase]):
        """Set the scheduler.

        Args:
            scheduler (Union[None, SchedulerBase]): The SchedulerBase derivative as
                the scheduler.
        """
        self._scheduler = self._validate_scheduler(scheduler)

    @property
    def to_tjp(self) -> str:
        """Convert the studio to a tjp representation.

        Returns:
            str: The TaskJuggler representation of this Studio.
        """
        start = time.time()

        tab = "    "
        indent = tab
        now = self.round_time(self.now).astimezone(pytz.utc).strftime("%Y-%m-%d-%H:%M")
        tjp = (
            f'project {self.tjp_id} "{self.tjp_id}" '
            f"{self.start.date()} - {self.end.date()} {{"
        )
        timing_resolution = (
            self.timing_resolution.days * 86400 + self.timing_resolution.seconds // 60
        )
        tjp += f"\n{indent}timingresolution {timing_resolution}min"
        tjp += f"\n{indent}now {now}"
        tjp += f"\n{indent}dailyworkinghours {self.daily_working_hours}"
        tjp += f"\n{indent}weekstartsmonday"

        # working hours
        tjp += "\n"
        tjp += "\n".join(
            f"{indent}{line}" for line in self.working_hours.to_tjp.split("\n")
        )

        tjp += f'\n{indent}timeformat "%Y-%m-%d"'
        tjp += f'\n{indent}scenario plan "Plan"'
        tjp += f"\n{indent}trackingscenario plan"
        tjp += "\n}"

        end = time.time()
        logger.debug("render studio to tjp took: {:0.3f} seconds".format(end - start))
        return tjp

    @property
    def projects(self) -> List[Project]:
        """Returns all the projects in the studio.

        Returns:
            List[Project]: List of all the Project instances in this Studio.
        """
        return Project.query.all()

    @property
    def active_projects(self) -> List[Project]:
        """Return all the active projects in the studio.

        Returns:
            List[Project]: List of active Project instances in this studio.
        """
        with DBSession.no_autoflush:
            wip = Status.query.filter_by(code="WIP").first()
        return Project.query.filter(Project.status == wip).all()

    @property
    def inactive_projects(self) -> List[Project]:
        """Return all the inactive projects in the studio.

        Returns:
            List[Project]: List of inactive Project instances in this studio.
        """
        with DBSession.no_autoflush:
            wip = Status.query.filter_by(code="WIP").first()
        return Project.query.filter(Project.status != wip).all()

    @property
    def departments(self) -> List[Department]:
        """Return all the departments in the studio.

        Returns:
            List[Department]: The list of Department instances in this Studio.
        """
        return Department.query.all()

    @property
    def users(self) -> List[User]:
        """Return all the users in the studio.

        Returns:
            List[User]: List of User instances in the studio.
        """
        return User.query.all()

    @property
    def vacations(self) -> List["Vacation"]:
        """Return all Vacations which doesn't have a User defined.

        Returns:
            List[Vacation]: List of Vacation instances.
        """
        return Vacation.query.filter(Vacation.user == None).all()  # noqa: E711

    def schedule(self, scheduled_by: Optional[User] = None) -> str:
        """Schedule all the active projects in the studio.

        Needs a Scheduler, so before calling it set a scheduler by using the
        :attr:`.scheduler` attribute.

        Args:
            scheduled_by (stalker.models.auth.User): A User instance who is doing the
                scheduling.

        Raises:
            RuntimeError: If the `self.scheduler` is None or it is not a `SchedulerBase`
                instance.

        Returns:
            str: The result of the scheduling process.
        """
        # check the scheduler first
        if self.scheduler is None or not isinstance(self.scheduler, SchedulerBase):
            raise RuntimeError(
                "There is no scheduler for this {cls}, please assign a scheduler to "
                "the {cls}.scheduler attribute, before calling {cls}.schedule()".format(
                    cls=self.__class__.__name__
                )
            )

        with DBSession.no_autoflush:
            self.scheduling_started_at = datetime.datetime.now(pytz.utc)

            # run the scheduler
            self.scheduler.studio = self
        start = time.time()

        # commit before scheduling
        # DBSession.commit()

        result = None
        try:
            result = self.scheduler.schedule()
        finally:
            # in any case set is_scheduling to False
            with DBSession.no_autoflush:
                self.is_scheduling = False
                self.is_scheduling_by = None

                # also store the result
                # if result:
                self.last_schedule_message = result

                # And the date the schedule is completed
                self.last_scheduled_at = datetime.datetime.now(pytz.utc)

                # and who has done the scheduling
                if scheduled_by:
                    logger.debug(f"setting last_scheduled_by to : {scheduled_by}")
                    self.last_scheduled_by = scheduled_by

        end = time.time()
        logger.debug("scheduling took {:0.3f} seconds".format(end - start))
        return result

    @property
    def weekly_working_hours(self) -> int:
        """Return the WorkingHours.weekly_working_hours value.

        Returns:
            int: The weekly working hours value stored in the working hours
                configuration of this Studio instance.
        """
        return self.working_hours.weekly_working_hours

    @property
    def weekly_working_days(self) -> int:
        """Return the WorkingHours.weekly_working_hours value.

        Returns:
            int: The weekly working days value stored in the working hours
                configuration of this Studio instance.
        """
        return self.working_hours.weekly_working_days

    @property
    def yearly_working_days(self) -> int:
        """Return the WorkingHours.yearly_working_days value.

        Returns:
            int: The yearly working days in the working hours configuration of this
                Studio instance.
        """
        return self.working_hours.yearly_working_days

    def to_unit(
        self,
        from_timing: int,
        from_unit: str,
        to_unit: str,
        working_hours: bool = True,
    ) -> int:
        """Convert the given timing and unit to the desired unit.

        If working_hours=True then the given timing is considered as working hours.

        Args:
            from_timing (int): The timing value.
            from_unit (str): The timing unit.
            to_unit (str): The other timing unit to convert the given timing unit to.
            working_hours (bool): True to consider the given from timing as a working
                hour. Default is True.

        Raises:
            NotImplementedError: Unless it is implemented.
        """
        raise NotImplementedError("this is not implemented yet")

    def _timing_resolution_getter(self) -> datetime.timedelta:
        """Return the timing_resolution value.

        Returns:
            datetime.timedelta: The timing resolution stored in this Studio instance.
        """
        return self._timing_resolution

    def _timing_resolution_setter(self, timing_resolution: datetime.timedelta) -> None:
        """Set the timing_resolution.

        Args:
            timing_resolution (datetime.timedelta): The `timing_resolution` instance to
                validate.
        """
        self._timing_resolution = self._validate_timing_resolution(timing_resolution)
        logger.debug(f"self._timing_resolution: {self._timing_resolution}")
        # update date values
        if self.start and self.end and self.duration:
            self._start, self._end, self._duration = self._validate_dates(
                self.round_time(self.start), self.round_time(self.end), None
            )

    timing_resolution: Mapped[Optional[datetime.timedelta]] = synonym(
        "_timing_resolution",
        descriptor=property(
            _timing_resolution_getter,
            _timing_resolution_setter,
            doc="""The timing_resolution of this object.

            Can be set to any value that is representable with
            datetime.timedelta. The default value is 1 hour. Whenever it is
            changed the start, end and duration values will be updated.
            """,
        ),
    )

    def _validate_timing_resolution(
        self, timing_resolution: datetime.timedelta
    ) -> datetime.timedelta:
        """Validate the given timing_resolution value.

        Args:
            timing_resolution (datetime.timedelta): The timing resolution value as a
                `datetime.timedelta` instance.

        Raises:
            TypeError: If the given `timing_resolution` is not a `datetime.timedelta`
                instance.

        Returns:
            datetime.timedelta: The validated timing resolution instance.
        """
        if timing_resolution is None:
            timing_resolution = defaults.timing_resolution

        if not isinstance(timing_resolution, datetime.timedelta):
            raise TypeError(
                "{}.timing_resolution should be an instance of "
                "datetime.timedelta, not {}: '{}'".format(
                    self.__class__.__name__,
                    timing_resolution.__class__.__name__,
                    timing_resolution,
                )
            )

        return timing_resolution


class WorkingHours(Entity):
    """A helper class to manage Studio working hours.

    Working hours is a data class to store the weekly working hours pattern of
    the studio.

    The data stored as a dictionary with the short day names are used as the
    key and the value is a list of two integers showing the working hours
    interval as the minutes after midnight. This is done in that way to ease
    the data transfer to TaskJuggler. The default value is defined in
    :class:`stalker.config.Config` ::

      wh = WorkingHours()
      wh.working_hours = {
          'mon': [[540, 720], [820, 1080]], # 9:00 - 12:00, 13:00 - 18:00
          'tue': [[540, 720], [820, 1080]], # 9:00 - 12:00, 13:00 - 18:00
          'wed': [[540, 720], [820, 1080]], # 9:00 - 12:00, 13:00 - 18:00
          'thu': [[540, 720], [820, 1080]], # 9:00 - 12:00, 13:00 - 18:00
          'fri': [[540, 720], [820, 1080]], # 9:00 - 12:00, 13:00 - 18:00
          'sat': [], # saturday off
          'sun': [], # sunday off
      }

    The default value is 9:00 - 18:00 from Monday to Friday and Saturday and
    Sunday are off.

    The working hours can be updated by the user supplied dictionary. If the
    user supplied dictionary doesn't have all the days then the default values
    will be used for those days.

    It is possible to use day index and day short names as a key value to reach
    the data::

      from stalker import config
      defaults = config.Config()

      wh = WorkingHours()

      # this is same by doing wh.working_hours['sun']
      assert wh['sun'] == defaults.working_hours['sun']

      # you can reach the data using the weekday number as index
      assert wh[0] == defaults.working_hours['mon']

      # working hours of sunday if defaults are used or any other day defined
      # by the stalker.config.Config.day_order
      assert wh[0] == defaults.working_hours[defaults.day_order[0]]

    Args:
        working_hours (Union[None, dict]): The dictionary that shows the
            working hours. The keys of the dictionary should be one of ['mon',
            'tue', 'wed', 'thu', 'fri', 'sat', 'sun']. And the values should be
            a list of two integers like [[int, int], [int, int], ...] format,
            showing the minutes after midnight. For missing days the default
            value will be used. If skipped the default value is going to be
            used.
        daily_working_hours (Union[None, int]): The daily working hours value.
            If given None the default value will be used.
    """

    __auto_name__ = True
    __tablename__ = "WorkingHours"
    __mapper_args__ = {"polymorphic_identity": "WorkingHours"}

    working_hours_id: Mapped[int] = mapped_column(
        "id", ForeignKey("Entities.id"), primary_key=True
    )

    working_hours: Mapped[Optional[Dict[str, List]]] = mapped_column(GenericJSON)
    daily_working_hours: Mapped[Optional[int]] = mapped_column(
        default=defaults.daily_working_hours
    )

    def __init__(
        self,
        working_hours: Optional[Dict[str, List]] = None,
        daily_working_hours=None,
        **kwargs: Dict[str, Any],
    ) -> None:
        super(WorkingHours, self).__init__(**kwargs)
        if working_hours is None:
            working_hours = defaults.working_hours
        self.working_hours = working_hours
        self.daily_working_hours = daily_working_hours

    def __eq__(self, other: Any) -> bool:
        """Check the equality.

        Args:
            other (Any): The other object.

        Returns:
            bool: True if the other object is a WorkingHours instance and has
                the same working_hours.
        """
        return (
            isinstance(other, WorkingHours)
            and other.working_hours == self.working_hours
        )

    def __getitem__(self, index: Union[int, str]) -> Optional[List]:
        """Return the item at the given index.

        Args:
            index (Union[int, str]): Either an integer representing the weekday starting
                from Monday:0 or a string value of a shorthand day name one of
                ["mon", "tue", "wed", "thu", "fri", "sat", "sun"].

        Returns:
            List[int, int]: The daily working hour arranged in a list where the first
                item is the minute from the midnight of the start of the working hour
                and the second item is the minute from the midnight of the end of the
                daily working hour. As in [540, 1080] represents 9am to 6pm.
        """
        if isinstance(index, int):
            return self.working_hours[defaults.day_order[index]]
        elif isinstance(index, str):
            return self.working_hours[index]

    def __setitem__(self, key: Union[int, str], value: List[List]) -> None:
        """Set the item value at the given index.

        Args:
            key (Union[int, str]): The index to set the value of or the day name.
            value (List[List[int, int]]): The working hours data arranged in a list of
                lists of two integers.

        Raises:
            KeyError: If the given key value is not one of the day names.
        """
        self._validate_working_hours_value(value)
        if isinstance(key, int):
            self.working_hours[defaults.day_order[key]] = value
        elif isinstance(key, str):
            # check if key is in
            if key not in defaults.day_order:
                raise KeyError(
                    "{} accepts only {} as key, not '{}'".format(
                        self.__class__.__name__, defaults.day_order, key
                    )
                )
            self.working_hours[key] = value

    @validates("working_hours")
    def _validate_working_hours(self, key: str, working_hours: Dict[str, List]) -> dict:
        """Validate the given working hours value.

        Args:
            key (str): The name of the validated column.
            working_hours (dict): The working hours value to be validated.

        Raises:
            TypeError: If the given working hours is not a dictionary.
            TypeError: If the values in the working hours dictionary are not lists.

        Returns:
            dict: The validated working hours dictionary.
        """
        if not isinstance(working_hours, dict):
            raise TypeError(
                "{}.working_hours should be a dictionary, not {}: '{}'".format(
                    self.__class__.__name__,
                    working_hours.__class__.__name__,
                    working_hours,
                )
            )

        for day in working_hours:
            if not isinstance(working_hours[day], list):
                raise TypeError(
                    '{}.working_hours should be a dictionary with keys "mon, '
                    'tue, wed, thu, fri, sat, sun" and the values should a '
                    "list of lists of two integers like [[540, 720], [800, "
                    "1080]], not {}: '{}'".format(
                        self.__class__.__name__,
                        working_hours[day].__class__.__name__,
                        working_hours[day],
                    )
                )

            # validate item values
            self._validate_working_hours_value(working_hours[day])

        # update the default values with the supplied working_hour dictionary
        # copy the defaults
        wh_def = copy.copy(defaults.working_hours)
        # update them
        wh_def.update(working_hours)

        return wh_def

    def is_working_hour(self, check_for_date: datetime.datetime) -> bool:
        """Check if the given datetime is in working hours.

        Args:
            check_for_date (datetime.datetime): The time value to check if it
                is a working hour.

        Returns:
            bool: True if the given datetime coincides to a working hour, False
                otherwise.
        """
        weekday_nr = check_for_date.weekday()
        hour = check_for_date.hour
        minute = check_for_date.minute

        time_from_midnight = hour * 60 + minute

        # check if the hour is inside the working hour ranges
        logger.debug(f"checking for: {time_from_midnight}")
        logger.debug(f"self[weekday_nr]: {self[weekday_nr]}")
        for working_hour_groups in self[weekday_nr]:
            start = working_hour_groups[0]
            end = working_hour_groups[1]
            logger.debug(f"start       : {start}")
            logger.debug(f"end         : {end}")
            if start <= time_from_midnight < end:
                return True

        return False

    def _validate_working_hours_value(self, value: List) -> List:
        """Validate the working hour value.

        The given value should follow the following format:

        .. code-block:: python

            working_hours = [
                [540, 1080],  # Working hour in minutes for Monday
                [540, 1080],  # Working hour in minutes for Tuesday
                [540, 1080],  # Working hour in minutes for Wednesday
                [540, 1080],  # Working hour in minutes for Thursday
                [540, 1080],  # Working hour in minutes for Friday
                [0, 0],  # Working hour in minutes for Saturday
                [0, 0],  # Working hour in minutes for Sunday
            ]

        Args:
            value (List): The validated working hour value.

        Raises:
            TypeError: If the given value is not a list.
            TypeError: If the immediate items in the list is not a list.
            TypeError: If the length of the items in the given list is not 2.
            TypeError: If the items in the lists inside the list are not integers.
            ValueError: If the integer values in the secondary lists are smaller than 0
                or larger than 1440 (which is 24 * 60).

        Returns:
            List[List[int, int]]
        """
        err = (
            "{}.working_hours value should be a list of lists of two "
            "integers and the range of integers should be between 0-1440, "
            "not {}: '{}'".format(
                self.__class__.__name__, value.__class__.__name__, value
            )
        )

        if not isinstance(value, list):
            raise TypeError(err)

        for i in value:
            if not isinstance(i, list):
                raise TypeError(err)

            # check list length
            if len(i) != 2:
                raise ValueError(err)

            # check type
            for j in range(2):
                if not isinstance(i[j], int):
                    raise TypeError(err)

                # check range
                if i[j] < 0 or i[j] > 1440:
                    raise ValueError(err)

        return value

    @property
    def to_tjp(self) -> str:
        """Return TaskJuggler representation of this object.

        Returns:
            str: The TaskJuggler representation.
        """
        tjp = ""
        for i, day in enumerate(["mon", "tue", "wed", "thu", "fri", "sat", "sun"]):
            if i != 0:
                tjp += "\n"
            tjp += f"workinghours {day} "
            if self[day]:
                for i, part in enumerate(self[day]):
                    start_hour, end_hour = part
                    if i != 0:
                        tjp += ", "
                    tjp += (
                        f"{start_hour // 60:02d}:{start_hour % 60:02d} - "
                        f"{end_hour // 60:02d}:{end_hour % 60:02d}"
                    )
            else:
                tjp += "off"
        return tjp

    @property
    def weekly_working_hours(self) -> int:
        """Return the total working hours in a week.

        Returns:
            int: The calculated weekly working hours.
        """
        weekly_working_hours = 0
        for i in range(0, 7):
            for start, end in self[i]:
                weekly_working_hours += end - start
        return weekly_working_hours / 60.0

    @property
    def weekly_working_days(self) -> int:
        """Return the weekly working days by looking at the working hours settings.

        Returns:
            int: The weekly working days value.
        """
        wwd = 0
        for i in range(0, 7):
            if len(self[i]):
                wwd += 1
        return wwd

    @property
    def yearly_working_days(self) -> int:
        """Return the total working days in a year.

        Returns:
            int: The calculated yearly_working_days value.
        """
        return int(ceil(self.weekly_working_days * 52.1428))

    @validates("daily_working_hours")
    def _validate_daily_working_hours(self, key: str, daily_working_hours: int) -> int:
        """Validate the given daily working hours value.

        Args:
            key (str): The name of the validated column.
            daily_working_hours (int): The daily working hours to be validated.

        Raises:
            TypeError: If the `daily_working_hours` value is not an integer.
            ValueError: If the `daily_working_hours` is smaller thane 0 or
                bigger than 24.

        Returns:
            int: The validated daily working hours value.
        """
        if daily_working_hours is None:
            daily_working_hours = defaults.daily_working_hours

        if not isinstance(daily_working_hours, int):
            raise TypeError(
                "{}.daily_working_hours should be an integer, not {}: '{}'".format(
                    self.__class__.__name__,
                    daily_working_hours.__class__.__name__,
                    daily_working_hours,
                )
            )

        if daily_working_hours <= 0 or daily_working_hours > 24:
            raise ValueError(
                f"{self.__class__.__name__}.daily_working_hours should be a positive "
                "integer value greater than 0 and smaller than or equal to 24"
            )
        return daily_working_hours

    def split_in_to_working_hours(
        self, start: datetime.datetime, end: datetime.datetime
    ) -> List[datetime.datetime]:
        """Split the given start and end datetime objects in to working hours.

        Args:
            start (datetime.datetime): The start date and time.
            end (datetime.datetime): The end date and time.

        Raises:
            NotImplementedError: Unless this is implemented.
        """
        raise NotImplementedError()


class Vacation(SimpleEntity, DateRangeMixin):
    """Vacation is the way to manage the User vacations.

    Args:
        user (User): The user of this vacation. Should be an instance of
            :class:`.User` if skipped or given as None the Vacation is considered
            as a Studio vacation and applies to all Users.

        start (datetime.datetime): The start datetime of the vacation. Is is an
            datetime.datetime instance. When skipped it will be set to the rounded
            value of.

        end (datetime.datetime): The end datetime of the vacation. It is an
            datetime.datetime instance.
    """

    __auto_name__ = True
    __tablename__ = "Vacations"
    __mapper_args__ = {"polymorphic_identity": "Vacation"}

    __strictly_typed__ = False

    vacation_id: Mapped[int] = mapped_column(
        "id", ForeignKey("SimpleEntities.id"), primary_key=True
    )

    user_id: Mapped[Optional[int]] = mapped_column("user_id", ForeignKey("Users.id"))

    user: Mapped[User] = relationship(
        primaryjoin="Vacations.c.user_id==Users.c.id",
        back_populates="vacations",
        doc="""The User of this Vacation.

        Accepts :class:`.User` instance.
        """,
    )

    def __init__(
        self,
        user: Optional[User] = None,
        start: Optional[datetime.datetime] = None,
        end: Optional[datetime.datetime] = None,
        **kwargs: Dict[str, Any],
    ) -> None:
        kwargs["start"] = start
        kwargs["end"] = end
        super(Vacation, self).__init__(**kwargs)
        DateRangeMixin.__init__(self, **kwargs)
        self.user = user

    @validates("user")
    def _validate_user(self, key: str, user: User) -> User:
        """Validate the given user instance.

        Args:
            key (str): The name of the validated column.
            user (User): The user value to be validated.

        Raises:
            TypeError: If the user value is not None and not a
                :class:`stalker.models.auth.User` instance.

        Returns:
            User: The validated user value.
        """
        if user is not None and not isinstance(user, User):
            raise TypeError(
                "{}.user should be an instance of stalker.models.auth.User, "
                "not {}: '{}'".format(
                    self.__class__.__name__, user.__class__.__name__, user
                )
            )
        return user

    @property
    def to_tjp(self) -> str:
        """Override the to_tjp method.

        Returns:
            str: The rendered tjp template.
        """
        tjp = "vacation "
        tjp += f"{self.start.astimezone(pytz.utc).strftime('%Y-%m-%d-%H:%M:%S')} - "
        tjp += f"{self.end.astimezone(pytz.utc).strftime('%Y-%m-%d-%H:%M:%S')}"
        return tjp



# -*- coding: utf-8 -*-
"""The declarative base class is situated here."""
import logging
from typing import Any, Type

from sqlalchemy.orm import declarative_base

from stalker.db.session import DBSession
from stalker.log import get_logger
from stalker.utils import make_plural

logger: logging.Logger = get_logger(__name__)


class ORMClass(object):
    """The base of the Base class."""

    query = DBSession.query_property()

    @property
    def plural_class_name(self) -> str:
        """Return plural name of this class.

        Returns:
            str: The plural version of this class.
        """
        return make_plural(self.__class__.__name__)


Base: Type[Any] = declarative_base(cls=ORMClass)



# -*- coding: utf-8 -*-
"""The venerable DBSession is situated here.

This is a runtime storage for the DB session. Greatly simplifying the usage of a
scoped session.
"""
from typing import Any, List, TYPE_CHECKING, Union

from sqlalchemy.orm import scoped_session, sessionmaker

if TYPE_CHECKING:  # pragma: no cover
    from stalker.models.entity import SimpleEntity


class ExtendedScopedSession(scoped_session):
    """A customized scoped_session which adds new functionality."""

    def save(self, data: Union[None, List[Any], "SimpleEntity"] = None) -> None:
        """Add and commits data at once.

        Args:
            data (Union[list, stalker.models.entity.SimpleEntity]): Either a single or
                a list of :class:`stalker.models.entity.SimpleEntity` or derivatives.
        """
        if data is not None:
            if isinstance(data, list):
                self.add_all(data)
            else:
                self.add(data)
        self.commit()


DBSession = ExtendedScopedSession(sessionmaker(future=True))



# -*- coding: utf-8 -*-
import os

# Local Imports
import stalker
from stalker import version


def test_version_number_is_correct():
    """version.VERSION is correct."""
    version_file_path = os.path.join(os.path.dirname(stalker.__file__), "VERSION")
    with open(version_file_path) as f:
        expected_version = f.read().strip()
    assert expected_version == version.__version__


def test_version_number_as_a_module_level_variable():
    """stalker.__version__ exists and value is correct."""
    assert version.__version__ == stalker.__version__



# -*- coding: utf-8 -*-
"""File related classes and utility functions are situated here."""

import os
from typing import Any, Dict, Generator, List, Optional, Union

from sqlalchemy import ForeignKey, String, Text
from sqlalchemy.orm import Mapped, mapped_column, validates

from stalker.log import get_logger
from stalker.models.entity import Entity
from stalker.models.enum import TraversalDirection
from stalker.models.mixins import ReferenceMixin
from stalker.utils import walk_hierarchy


logger = get_logger(__name__)


class File(Entity, ReferenceMixin):
    """Holds data about files or file sequences.

    Files are all about giving some external information to the current entity
    (external to the database, so it can be something on the
    :class:`.Repository` or in the Web or anywhere that the server can reach).
    The type of the file (general, file, folder, web page, image, image
    sequence, video, movie, sound, text etc.) can be defined by a
    :class:`.Type` instance (you can also use multiple :class:`.Tag` instances
    to add more information, and to filter them back). Again it is defined by
    the needs of the studio.

    For sequences of files the file name should be in "%h%p%t %R" format in
    PySeq_ formatting rules.

    There are three secondary attributes (properties to be more precise)
    ``path``, ``filename`` and ``extension``. These attributes are derived from
    the :attr:`.full_path` attribute and they modify it.

    Path
        It is the path part of the full_path.

    Filename
        It is the filename part of the full_path, also includes the extension,
        so changing the filename also changes the extension part.

    Extension
        It is the extension part of the full_path. It also includes the
        extension separator ('.' for most of the file systems).

    .. versionadded:: 1.1.0

       Inputs or references can now be tracked per File instance through the
       :attr:`.File.references` attribute. So, that all the references can be
       tracked per individual file instance.

    Args:
        full_path (str): The full path to the File, it can be a path to a
            folder or a file in the file system, or a web page. For file
            sequences use "%h%p%t %R" format, for more information see
            `PySeq Documentation`_. It can be set to empty string (or None
            which will be converted to an empty string automatically).

    .. _PySeq: http://packages.python.org/pyseq/
    .. _PySeq Documentation: http://packages.python.org/pyseq/
    """

    __auto_name__ = True
    __tablename__ = "Files"
    __mapper_args__ = {"polymorphic_identity": "File"}

    file_id: Mapped[int] = mapped_column(
        "id",
        ForeignKey("Entities.id"),
        primary_key=True,
    )

    # this is a limit for most
    original_filename: Mapped[Optional[str]] = mapped_column(String(256))
    # file systems
    full_path: Mapped[Optional[str]] = mapped_column(
        Text, doc="The full path of the url to the file."
    )

    created_with: Mapped[Optional[str]] = mapped_column(String(256))

    def __init__(
        self,
        full_path: Optional[str] = "",
        original_filename: Optional[str] = "",
        references: Optional[List["File"]] = None,
        created_with: Optional[str] = None,
        **kwargs: Optional[Dict[str, Any]],
    ) -> None:
        super(File, self).__init__(**kwargs)
        ReferenceMixin.__init__(self, references=references)
        self.full_path = full_path
        self.original_filename = original_filename
        self.created_with = created_with

    @validates("full_path")
    def _validate_full_path(self, key: str, full_path: Union[None, str]) -> str:
        """Validate the given full_path value.

        Args:
            key (str): The name of the validated column.
            full_path (str): The full_path value to be validated.

        Raises:
            TypeError: If the given full_path is not a str.

        Returns:
            str: The validated full_path value.
        """
        if full_path is None:
            full_path = ""

        if not isinstance(full_path, str):
            raise TypeError(
                f"{self.__class__.__name__}.full_path should be a str, "
                f"not {full_path.__class__.__name__}: '{full_path}'"
            )

        return self._format_path(full_path)

    @validates("created_with")
    def _validate_created_with(
        self, key: str, created_with: Union[None, str]
    ) -> Union[None, str]:
        """Validate the given created_with value.

        Args:
            key (str): The name of the validated column.
            created_with (str): The name of the application used to create this
                File.

        Raises:
            TypeError: If the given created_with attribute is not None and not
                a string.

        Returns:
            Union[None, str]: The validated created with value.
        """
        if created_with is not None and not isinstance(created_with, str):
            raise TypeError(
                "{}.created_with should be an instance of str, not {}: '{}'".format(
                    self.__class__.__name__,
                    created_with.__class__.__name__,
                    created_with,
                )
            )
        return created_with

    @validates("original_filename")
    def _validate_original_filename(
        self, key: str, original_filename: Union[None, str]
    ) -> str:
        """Validate the given original_filename value.

        Args:
            key (str): The name of the validated column.
            original_filename (str): The original filename value to be validated.

        Raises:
            TypeError: If the given original_filename value is not a str.

        Returns:
            str: The validated original_filename value.
        """
        filename_from_path = os.path.basename(self.full_path)
        if original_filename is None:
            original_filename = filename_from_path

        if original_filename == "":
            original_filename = filename_from_path

        if not isinstance(original_filename, str):
            raise TypeError(
                f"{self.__class__.__name__}.original_filename should be a str, "
                f"not {original_filename.__class__.__name__}: '{original_filename}'"
            )

        return original_filename

    @staticmethod
    def _format_path(path: Union[bytes, str]) -> str:
        """Format the path to internal format.

        The path is using the Linux forward slashes for path separation.

        Args:
            path (Union[bytes, str]): The path value to be formatted.

        Returns:
            str: The formatted path value.
        """
        if isinstance(path, bytes):
            path = path.decode("utf-8")

        return path.replace("\\", "/")

    @property
    def path(self) -> str:
        """Return the path part of the full_path.

        Returns:
            str: The path part of the full_path value.
        """
        return os.path.split(self.full_path)[0]

    @path.setter
    def path(self, path: str) -> None:
        """Set the path part of the full_path attribute.

        Args:
            path (str): The new path value.

        Raises:
            TypeError: If the given path value is not a str.
            ValueError: If the given path is an empty str.
        """
        if path is None:
            raise TypeError(f"{self.__class__.__name__}.path cannot be set to None")

        if not isinstance(path, str):
            raise TypeError(
                f"{self.__class__.__name__}.path should be a str, "
                f"not {path.__class__.__name__}: '{path}'"
            )

        if path == "":
            raise ValueError(
                f"{self.__class__.__name__}.path cannot be an empty string"
            )

        self.full_path = self._format_path(os.path.join(path, self.filename))

    @property
    def filename(self) -> str:
        """Return the filename part of the full_path attribute.

        Returns:
            str: The filename part of the full_path attribute.
        """
        return os.path.split(self.full_path)[1]

    @filename.setter
    def filename(self, filename: Union[None, str]) -> None:
        """Set the filename part of the full_path attr.

        Args:
            filename (Union[None, str]): The new filename.

        Raises:
            TypeError: If the given filename is not a str.
        """
        if filename is None:
            filename = ""

        if not isinstance(filename, str):
            raise TypeError(
                f"{self.__class__.__name__}.filename should be a str, "
                f"not {filename.__class__.__name__}: '{filename}'"
            )

        self.full_path = self._format_path(os.path.join(self.path, filename))

    @property
    def extension(self) -> str:
        """Return the extension value.

        Returns:
            str: The extension extracted from the full_path value.
        """
        return os.path.splitext(self.full_path)[1]

    @extension.setter
    def extension(self, extension: Union[None, str]) -> None:
        """Set the extension value.

        Args:
            extension (Union[None, str]): The new extension value.

        Raises:
            TypeError: If the given extension value is not a str.
        """
        if extension is None:
            extension = ""

        if not isinstance(extension, str):
            raise TypeError(
                f"{self.__class__.__name__}.extension should be a str, "
                f"not {extension.__class__.__name__}: '{extension}'"
            )

        if extension != "":
            if not extension.startswith(os.path.extsep):
                extension = os.path.extsep + extension

        self.filename = os.path.splitext(self.filename)[0] + extension

    @property
    def absolute_full_path(self) -> str:
        """Return the absolute full path of the file.

        Returns:
            str: The absolute full path of the file.
        """
        return os.path.normpath(os.path.expandvars(self.full_path))

    @property
    def absolute_path(self) -> str:
        """Return the absolute path of the file.

        Returns:
            str: The absolute path of the file.
        """
        return os.path.dirname(self.absolute_full_path)

    def walk_references(
        self,
        method: Union[int, str, TraversalDirection] = TraversalDirection.DepthFirst,
    ) -> Generator[None, "File", None]:
        """Walk the references of this file.

        Args:
            method (Union[int, str, TraversalDirection]): The walk method
                defined by the :class:`.TraversalDirection` enum.

        Yields:
            File: Yield the File instances.
        """
        for v in walk_hierarchy(self, "references", method=method):
            yield v

    def __eq__(self, other: Any) -> bool:
        """Check if the other is equal to this File.

        Args:
            other (Any): The other object to be checked for equality.

        Returns:
            bool: If the other object is a File instance and has the same
                full_path and type value.
        """
        return (
            super(File, self).__eq__(other)
            and isinstance(other, File)
            and self.full_path == other.full_path
            and self.type == other.type
        )

    def __hash__(self) -> int:
        """Return the hash value of this instance.

        Because the __eq__ is overridden the __hash__ also needs to be overridden.

        Returns:
            int: The hash value.
        """
        return super(File, self).__hash__()
